unit BaseUtils;
{==================================================================

 Basic file handling utilities  /  Version 1.1  / 5.1999
 ========================================================


    Description:
    ============
    Basic utilities useful for handling files or directories.
    Used by the components TDriveView and TDirView.

    Author:
    =======
    (c) Ingo Eckel 5/1998
    Sodener Weg 38
    65812 Bad Soden
    Germany

{==================================================================}

interface

Uses SysUtils, Windows;


Function  LastPos(Ch : Char; S : String)               : Integer;
Function  StrReplace(S : String; FromC, ToC : Char)    : String;
Procedure StrTranslate(Var S: String; Code : String);
Function  StrContains(Str1, Str2 : String)             : Boolean;
Function  StrDeleteChar(S : String; Ch : Char)         : String;
Function  FileOrDirExists(FileName : String)           : Boolean;
Function  DirExists(Dir : String)                      : Boolean; OverLoad;
Function  DirExists(Dir : String; Var Attrs : Integer) : Boolean; OverLoad;
Function  AddSlash(S : String)                         : String;
Function  RemoveSlash(S : String)                      : String;
function  DiskInDrive(Drive: Char)                     : Boolean;
Function  ExtractFileNameOnly(Name : String)           : String;
Function  ExtractLastDirName(Dir : String)             : String;
Function  DosDateToDateStr(wDate : Word)               : String;
Function  StrDateToDosDate(DateStr : String)           : Word;
Function  FormatSize(Size : Integer)                  : String; Overload;
Function  FormatSize(Size : Cardinal)                  : String; Overload;
Function  FormatSize(Size : Int64)                     : String; Overload;

Implementation


Function LastPos(Ch : Char; S : String) : Integer;
Var i : Integer;
Begin
  Result := 0;
  For i := Length(S) DownTo 1 do
  IF S[i] = Ch Then
  Begin
    Result := i;
    Exit;
  End;
End; {LastPos}




Function StrReplace(S : String; FromC, ToC : Char) : String;
Var FoundPos : Integer;
Begin
  FoundPos := Pos(FromC, S);
  While FoundPos <> 0 Do
  Begin
    S[FoundPos] := ToC;
    FoundPos := Pos(FromC, S);
  End;
  Result := S;
End;

Procedure StrTranslate(Var S: String; Code : String);
Var i : Integer;
Begin
  IF (Length(Code) = 0) Or (Length(Code) Mod 2 <> 0) Then Exit;
  For i := 1 To Length(Code) Div 2 Do
  Begin
    While Pos(Code[i*2-1], S) <> 0 Do
    S[Pos(Code[i*2-1], S)] := Code[i*2];
  End;
End; {Translate}


Function StrDeleteChar(S : String; Ch : Char) : String;
Var chPos : Integer;
Begin
  chPos := Pos(Ch, S);
  WHile chPos > 0 Do
  Begin
    Delete(S, chPos, 1);
    chPos := Pos(Ch, S);
  End;
  Result := S;
End; {StrDelChar}


Function StrContains(Str1, Str2 : String) : Boolean;
Var i : Integer;
Begin
  For i := 1 To Length(Str1) Do
    IF Pos(Str1[i], Str2) <> 0 Then
    Begin
      Result := True;
      Exit;
    End;
  Result := False;
End; {StringCountains}


Function FileOrDirExists(FileName : String) : Boolean;
Var SRec : TSearchRec;
Begin
  Result := FindFirst(FileName, faAnyFile, SRec) = 0;
  SysUtils.FindCLose(SRec);
End; {FileOrDirExists}


Function DirExists(Dir : String; Var Attrs : Integer) : Boolean;
Var SRec : TSearchRec;
Begin
  Result := ((Length(Dir) <= 3) And (Length(Dir) >= 2))And (Dir[2] = ':');
  IF Result Then
   Attrs := 0
  Else
  Begin
    IF FindFirst(Dir, faAnyFile, SRec) = 0 Then
    Begin
      Result := (Srec.Attr and faDirectory <> 0);
      Attrs := Srec.Attr;
    End;
    SysUtils.FindCLose(SRec);
  End;
End; {DirExists}


Function DirExists(Dir : String) : Boolean;
Var Dummy : Integer;
Begin
  Result := DirExists(Dir, Dummy);
End; {DirExists}


Function AddSlash(S : String) : String;
Begin
  IF (Length(S) > 0) And (S[Length(s)] <> '\') Then
  Result := S + '\'
  else
  Result := S;
End; {AddSlash}



Function RemoveSlash(S : String) : String;
Begin
  IF (Length(S) > 0) And (S[Length(s)] = '\') Then
  Result := Copy(S, 1, Pred(Length(S)))
  else
  Result := S;
End; {RemoveSlash}


function DiskInDrive(Drive: Char): Boolean;
var  ErrorMode: word;
begin
  { make it upper case }
  Drive := UpCase(Drive);
  { make sure it's a letter }
  if not (Drive in ['A'..'Z']) then
    raise EConvertError.Create('Not a valid drive ID');
  { turn off critical errors }
  ErrorMode := SetErrorMode(SEM_FailCriticalErrors or SEM_NOOPENFILEERRORBOX);
  try
    { drive 1 = a, 2 = b, 3 = c, etc. }
    if DiskSize(Ord(Drive) - $40) = -1 then
    Result := False
    else
      Result := True;
  finally    { restore old error mode }
    SetErrorMode(ErrorMode);
  end;
end; {DiskInDrive}


Function ExtractFileNameOnly(Name : String) : String;
Var Ext : String;
Begin
  Result := ExtractFileName(Name);
  Ext    := ExtractFileExt(Name);
  IF Ext <> '' Then
  Delete(Result, Pos(Ext, Result), Length(Ext));
End; {ExtractFileNameOnly}


Function ExtractLastDirName(Dir : String) : String;
Var SPos : Integer;
Begin
  Result := RemoveSlash(Dir);
  SPos   := LastPos('\', Result);
  IF SPos = 0 Then
  Result := ''
  Else
  Result := Copy(Result, Succ(SPos), Length(Result) - SPos);
End; {ExtractLastDirName}

Function DosDateToDateStr(wDate : Word) : String;
Var FileTime    : TFileTime;
    SysTime     : TSystemTime;
    DateTime    : TDateTime;

Begin
    DosDateTimeToFileTime(wDate, 0, FileTime);
    FileTimeToSystemTime(FileTime, SysTime);
    DateTime := SystemTimeToDateTime(SysTime);
    Result := FormatDateTime(ShortDateFormat, DateTime);
End; {DosDateTimeToStr}



Function StrDateToDosDate(DateStr : String) : Word;
Var FileTime    : TFileTime;
    SysTime     : TSystemTime;
    DateTime    : TDateTime;
    wTime       : Word;
    wDate       : Word;

Begin
  DateTime := StrToDate(DateStr);
  DateTimeToSystemTime(DateTime, SysTime);
  SystemTimeToFileTime(SysTime, FileTime);
  FileTimeToDosDateTime(FileTime, wDate , wTime);
  Result := wDate;
End; {StrToDosDateTime}


Function  FormatSize(Size : Integer) : String;
Begin
  Result := FormatSize(Cardinal(ABS(Size)));
End; {FormatSize}


Function FormatSize(Size : Cardinal) : String;
Var i : Integer;
    p : Integer;

Begin
  p := 0;
  i := 3;
  Result := IntToStr(Size);
  While i + p < Length(Result) Do
  Begin
    Insert(ThousandSeparator, Result, Length(Result) - (i + p)+ 1);
    Inc(p);
    INC(i, 3);
  End;
End; {FormatSize}



Function FormatSize(Size : Int64) : String;
Var i : Integer;
    p : Integer;

Begin
  p := 0;
  i := 3;
  Result := IntToStr(Size);
  While i + p < Length(Result) Do
  Begin
    Insert(ThousandSeparator, Result, Length(Result) - (i + p)+ 1);
    Inc(p);
    INC(i, 3);
  End;
End; {FormatSize}

end.
