Unit DriveView;

{==================================================================
 Component TDriveView / Version 2.4, July 1999
 ==================================================================


    Description:
    ============
    Displays the the directory structure of all drives as treeview
    with shell icons. Complete drag&Drop support for files and
    directories.


    Author:
    =======
    (c) Ingo Eckel 1998, 1999
    Sodener Weg 38
    65812 Bad Soden
    Germany

    For detailed documentation and history see TDriveView.htm.

    V2.4:
    - Don't accept files dropped from the recycle bin to the recycle bin.


{==================================================================}
{ Define DRAGDROP to enable drag&drop for directories:}
{$DEFINE DRAGDROP}


{Required compiler options for TDriveView:}
{$A+,B-,X+,H+,P+}
interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  ComCtrls,
  ShellApi,
  CommCtrl,
  ExtCtrls,
  ActiveX,
  Math,
  ShlObj,
  DirView,
  ShellDialogs,
{$IFDEF DRAGDROP}
  DragDrop,
  DragDropFilesEx,
  FileChanges,
{$ENDIF}
  FileOperator,
  DiscMon,
  IEDriveInfo,
  PIDL,
  BaseUtils,
  ListExt;


{$I ResStrings.pas }



Const coLogFileID         ='TDIRVIEW';
      coLogFileVersion    = '1.0';
      msThreadChangeDelay = 50;
      InvalidSize         = $FFFFFFFF;

{$IFDEF DRAGDROP}
      DDVScrollDelay      = 2000000;
      DDHScrollDelay      = 2000000;
      DDDragStartDelay    = 2000000;
      DDExpandDelay       = 25000000;
{$ENDIF}


      {Flags used by TDriveView.RefreshRootNodes:}
      dvdsFloppy          = 8;  {Include floppy drives}
      dvdsRereadAllways   = 16; {Refresh drivestatus in any case}

      {Types uses by the function IterateSubTree:}
      {TRecursiveScan: determines, wich nodes are scanned by the function IterateSubTree:
      rsNoRecursive:       Scan startnode only.
      rsRecursive:         Scan all subnodes of the startnode.
      rsRecursiveExisting: Scan all subnodes of the startnode but not new created subnodes.
      rsRecursiveExpanded: Scan all expanded subnodes of the startnode.}
      {TScanStartnode: determines, wether the startnode should also be scanned:}

Type TRecursiveScan      = (rsNoRecursive, rsRecursive, rsRecursiveExisting, rsRecursiveExpanded);
     TScanStartNode      = (coNoScanStartNode, coScanStartNode);

     TVolumeDisplayStyle = (doPrettyName, doDisplayName, doLongPrettyName); {Diplaytext of drive node}

     TString12 = String[12];


     TCallBackFunc = Function(Var Node : TTreeNode; Data: Pointer) : Boolean of Object;


  ECreateShortCut  = Class(Exception);
  EInvalidDirName  = Class(Exception);
  EInvalidPath     = Class(Exception);
  ENodeNotAssigned = Class(Exception);

  TDriveStatus = Record
    Scanned     : Boolean;     {Drive allready scanned?}
    Verified    : Boolean;     {Drive completly scanned?}
    RootNode    : TTreeNode;   {Rootnode to drive}
    DiscMonitor : TDiscMonitor;{Monitor thread}
    ChangeTimer : TTimer;      {Change timer for the monitor thread}
  End;


  TScanDirInfo = Record
    StartNode : TTreeNode;
    NewDirFound : Boolean;
  End;

  PScanDirInfo = ^TScanDirInfo;


  TDriveViewScanDirEvent    = Procedure(Sender: TObject; Node: TTreeNode; Var DoScanDir : Boolean) of object;
  TDriveViewDiskChangeEvent = Procedure(Sender: TObject; Drive : TDrive) of object;


  TNodeData = Class
    Private
      fDirName  : String;
      fShortName: TString12;
      fAttr     : Integer;
      fVerified : Boolean;
      fData     : Pointer;
      fExpanded : Boolean;
      fDrawBold : Boolean;
      fNewDir   : Boolean;
      fDirSize  : Cardinal;
      fisRecycleBin : Boolean;

    Public
      Property DirName      : String    Read fDirName   Write fDirName;
      Property ShortName    : TString12 Read fShortName Write fShortName;
      Property Attr         : Integer   Read Fattr      Write Fattr;
      Property Verified     : Boolean   Read FVerified  Write fVerified;
      Property Data         : Pointer   Read fData      Write fData;
      Property Expanded     : Boolean   Read fExpanded  Write fExpanded;
      Property DrawBold     : Boolean   Read fDrawBold  Write fDrawBold;
      Property NewDir       : Boolean   Read fNewDir    Write fNewDir;
      Property DirSize      : Cardinal  Read fDirSize   Write fDirSize;
      Property isRecycleBin : Boolean   Read fIsRecycleBin;
      Constructor Create;
      Destructor Destroy; Override;
        End;


{---------------------------------------------------------------}
  TDriveView = class(TCustomTreeView)
{---------------------------------------------------------------}
  private
{---------------------------------------------------------------}
    DriveStatus     : Array[FirstDrive .. LastDrive] Of TDriveStatus;

    fConfirmDelete    : Boolean;
    fConfirmOverwrite : Boolean;
    fWatchDirectory   : Boolean;
    FDirectory        : String;
    fFullDriveScan    : Boolean;
    fDimmHiddenDirs   : Boolean;
    fColorBold        : TColor;
    fShowDirSize      : Boolean;
    fShowVolLabel     : Boolean;
    fVolDisplayStyle  : TVolumeDisplayStyle;
    FUseSystemContextMenu : Boolean;
    FContinue         : Boolean;
    FShowAnimation    : Boolean;

    fChangeFlag       : Boolean;
    fContextMenu      : Boolean;
    fLastDir          : String;
    fValidateFlag     : Boolean;
    fCreating         : Boolean;
    FParentForm       : TCustomForm;

    {Additional events:}
    fOnStartScan      : TNotifyEvent;
    fOnEndScan        : TNotifyEvent;
    fOnScanDir        : TDriveViewScanDirEvent;
    fOnDiskChange     : TDriveViewDiskChangeEvent;
    fOnChangeDetected : TDriveViewDiskChangeEvent;
    fOnChangeInvalid  : TDriveViewDiskChangeEvent;

    {used components:}
    fDirView          : TDirView;
    FDriveBox         : TObject;
    FFileOperator     : TFileOperator;

    fChangeInterval   : Cardinal;
    FCanChange        : Boolean;

{$IFDEF DRAGDROP}
    {Drag&drop:}
    FDragDrive              : TDrive;
    DragFileList            : TStringList;
    DragNode                : TTreeNode;
    FDD                     : TDragDropFilesEx;
    DragOverTime            : FILETIME;
    DragStartTime           : FILETIME;
    LastVScrollTime         : FILETIME;
    LastHScrollTime         : FILETIME;
    VScrollCount            : Integer;
    FLastPathCut            : String;
    FTargetPopUpMenu        : Boolean;
    FUseDragImages          : Boolean;
    FStartPos               : TPoint;
    FDragPos                : TPoint;
    FExeDrag                : Boolean;

    FOnDDDragEnter          : TDDOnDragEnter;
    FOnDDDragLeave          : TDDOnDragLeave;
    FOnDDDragOver           : TDDOnDragOver;
    FOnDDDrop               : TDDOnDrop;
    FOnDDQueryContinueDrag  : TDDOnQueryContinueDrag;
    FOnDDGiveFeedback       : TDDOnGiveFeedback;
    FOnDDDragDetect         : TDDOnDragDetect;
    FOnDDProcessDropped     : TOnProcessDropped;
    FOnDDError              : TDDErrorEvent;
    FOnDDExecuted           : TDDExecutedEvent;
    FOnDDFileOperation      : TDDFileOperationEvent;
    FOnDDFileOperationExecuted : TDDFileOperationExecutedEvent;

    {Drag&Drop eventhandling:}
    Procedure DDDragEnter(DataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: longint; var Accept:boolean);
    Procedure DDDragLeave;
    Procedure DDDragOver(grfKeyState: Longint; pt: TPoint; var dwEffect: longint);
    Procedure DDDrop(DataObj: IDataObject; grfKeyState: Longint;  pt: TPoint; var dwEffect: longint);
    Procedure DDQueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint; var Result: HResult);
    Procedure DDGiveFeedback(dwEffect: Longint; var Result: HResult);
    Procedure DDDragDetect(grfKeyState: Longint; DetectStart, Pt: TPoint; DragStatus:TDragDetectStatus);
    Procedure DDProcessDropped(Sender: TObject; grfKeyState: Longint;  pt: TPoint; dwEffect: Longint);

    Procedure SetTargetPopUpMenu(PopMe : Boolean);

    {Drag&drop helper functions:}
    Procedure SignalDirDelete(Sender: TObject; Files : TStringList);
    Procedure PerformDragDropFileOperation(TargetPath : String; dwEffect: Integer; isRecycleBin : Boolean);
{$ENDIF}


    {Callback-functions used by iteratesubtree:}
    Function  CallBackValidateDir      (Var Node : TTreeNode; Data: Pointer) : Boolean;
    Function  CallBackSaveNode         (Var Node : TTreeNode; Data: Pointer) : Boolean;
    Function  CallBackSaveNodeState    (Var Node : TTreeNode; Data: Pointer) : Boolean;
    Function  CallBackRestoreNodeState (Var Node : TTreeNode; Data: Pointer) : Boolean;
    Function  CallBackDisplayName      (Var Node : TTreeNode; Data: Pointer) : Boolean;
    Function  CallBackSetDirSize       (Var Node : TTreeNode; Data: Pointer) : Boolean;
    Function  CallBackExpandLevel      (Var Node : TTreeNode; Data: Pointer) : Boolean;

    {Notification procedures used by component TDiscMonitor:}
    Procedure ChangeDetected(Sender: TObject);
    Procedure ChangeInvalid(Sender: TObject);

    {Notification procedure used by component TTimer:}
    Procedure ChangeTimerOnTimer(Sender : TObject);

    {Special procedure for events OnEdited / OnDrawItem. Used to overwrite these events:}
    Procedure InternalOnDrawItem(Sender: TCustomTreeView; Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);

{---------------------------------------------------------------}
  protected
{---------------------------------------------------------------}
    Procedure SetSelected(Node : TTreeNode);
    Procedure SetFullDriveScan(DoFullDriveScan : Boolean);
    Procedure SetWatchDirectory(Watch : Boolean);
    Procedure SetShowDirSize(ShowIt : Boolean);
    Procedure SetShowVolLabel(ShowIt : Boolean);
    Procedure SetVolDisplayStyle(doStyle : TVolumeDisplayStyle);
    Procedure SetDirView(DV : TDirView);
    Procedure SetChangeInterval(Interval : Cardinal);
    Procedure SetDimmHiddenDirs(DimmIt : Boolean);
    Function  GetDirectory : String;
    Procedure SetDirectory(Path : String);
    Function  GetNodeFromHItem(Item: TTVItem): TTreeNode;
    Function  DoScanDir(FromNode : TTreeNode) : Boolean;                           Virtual;
    Function  AddChildNode(ParentNode : TTreeNode; SRec : TSearchRec) : TTreeNode; Virtual;
    Procedure CreateWatchThread(Drive : TDrive);                                   Virtual;

{$IFDEF DRAGDROP}
    Procedure SetLastPathCut(Path : String);
    Function  GetCanUndoCopyMove : Boolean;                 Virtual;
    Procedure DDError(ErrorNo : TDDError);                  Dynamic;
    Procedure CNNotify(Var Msg: TWMNotify);                 Message CN_NOTIFY;
{$ENDIF}
    Procedure CreateWnd;                                    Override;
    Procedure Edit(Const Item: TTVItem);                    Override;
    Procedure Notification(AComponent: TComponent; Operation: TOperation); Override;

    Procedure WMLButtonDown(var Msg: TWMLButtonDown);       Message WM_LBUTTONDOWN;
    Procedure WMLButtonUp  (var Msg: TWMLButtonDown);       Message WM_LBUTTONUP;
    Procedure WMRButtonDown(Var Msg: TWMRButtonDown);       Message WM_RBUTTONDOWN;
    Procedure WMRButtonUp  (Var Msg: TWMRButtonUp);         Message WM_RBUTTONUP;
{---------------------------------------------------------------}
  public
{---------------------------------------------------------------}
    {Runtime-only properties:}

    property  Images;
    Property  StateImages;
    Property  Items Stored False;
    Property  Selected Write SetSelected;
    Property  Continue : Boolean Read FContinue
                                 Write FContinue;

    Property DriveBox : TObject Read FDriveBox
                                Write FDriveBox;

{$IFDEF DRAGDROP}
    Property DragDropFilesEx : TDragDropFilesEx Read FDD;
    Property DragDrive       : TDrive Read FDragDrive;
    Property CanUndoCopyMove : Boolean Read GetCanUndoCopyMove;
    Property DDFileOperator  : TFileOperator Read FFileOperator;
    Property LastPathCut     : String        Read FLastPathCut
                                             Write SetLastPathCut;

    Function  UndoCopyMove : Boolean;                          Dynamic;
    Procedure EmptyClipboard;                                  Dynamic;
    Function  CopyToClipBoard(Node : TTreeNode) : Boolean;     Dynamic;
    Function  CutToClipBoard(Node : TTreeNode)  : Boolean;     Dynamic;
    Function  CanPasteFromClipBoard : Boolean;                 Dynamic;
    Function  PasteFromClipBoard : Boolean;                    Dynamic;
{$ENDIF}

    {Drive handling:}
    Function  GetDriveStatus(Drive : TDrive) : TDriveStatus;
    Function  GetDriveTypetoNode(Node : TTreeNode) : Integer;  {Returns DRIVE_CDROM etc..}
    Function  GetDriveType(Drive : TDrive) : Integer;           {Returns DRIVE_CDROM etc..}
    Function  GetDrive(Node : TTreeNode) : Char;
    Function  GetDriveText(Drive : TDrive) : String;
    Procedure ScanDrive(Drive : TDrive);
    Procedure RefreshRootNodes(ScanDirectory : Boolean; dsFlags : Integer);
    Function  GetValidDrivesStr : String;
    Procedure RefreshDirSize(Node : TTreeNode);
    Procedure RefreshDriveDirSize(Drive : TDrive);

    {Node handling:}
    Procedure SetImageIndex(Node : TTreeNode);                                     Virtual;
    Function  FindNodeToPath(Path : String) : TTreeNode;                           Virtual;
    Procedure SetBoldDraw(Node : TTreeNode; BoldDraw : Boolean);                   Dynamic;
    Function  NodeVerified(Node : TTreeNode) : Boolean;                            Virtual;
    Function  NodeAttr(Node : TTreeNode) : Integer;                                Virtual;
    Function  RootNode(Node : TTreeNode) : TTreeNode;                              Virtual;
    Function  GetDirPathName(Node: TTreeNode) : String;                            Virtual;
    Function  GetDirPath (Node : TTreeNode) : String;                              Virtual;
    Function  GetDirName(Node : TTreeNode)   : String;                             Virtual;
    Procedure CenterNode(Node : TTreeNode);                                        Virtual;
    Function  SortChildren(ParentNode : TTreeNode; Recurse : Boolean) : Boolean;   Virtual;
    Function  GetDirSize(Node : TTreeNode) : Cardinal;                             Virtual;
    Procedure SetDirSize(Node : TTreeNode);                                        Virtual;
    Function  GetDisplayName(Node : TTreeNode) : String;                           Virtual;
    Function  NodeUpdateAble(Node : TTreeNode) : Boolean;                          Virtual;
    Function  FormatDirSize(Size : Cardinal) : String;                             Virtual;
    Procedure ExpandLevel(Node : TTreeNode; Level : Integer);                      Virtual;


    Procedure ValidateDirectoryEx(Node : TTreeNode; Recurse : TRecursiveScan);     Virtual;
    Procedure ValidateDirectory(Node : TTreeNode);                                 Virtual;
    Procedure ValidateDirectoryEasy(Node : TTreeNode);                             Virtual;
    Procedure ValidateExistingDirectories(Node : TTreeNode);                       Virtual;
    Procedure ValidateAllDirectories(Node : TTreeNode);                            Dynamic;
    Function  GetSubTreeSize(Node : TTreeNode) : Integer;                          Dynamic;

    {Directory update:}
    Function  CreateDirectory(ParentNode : TTreeNode; NewName : String) : TTreeNode; Dynamic;
    Function  DeleteDirectory(Node: TTreeNode; AllowUndo : Boolean)     : Boolean;   Dynamic;

    Procedure DeleteSubNodes(Node : TTreeNode);                                      Dynamic;

    {Basic recursive function for scanning a subtree:}
    Function  IterateSubTree(Var StartNode : TTreeNode;
                             CallBackFunc  : TCallBackFunc;
                             Recurse       : TRecursiveScan;
                             ScanStartNode : TScanStartNode;
                             Data          : Pointer) : Boolean;

    constructor Create(AOwner: TComponent);           Override;
    Destructor  Destroy;                              Override;

    {Load/save subtree nodes from/to file:}
    Function LoadSubNodesFromFile(ParentNode : TTreeNode; FileName : String) : Boolean;
    Function SaveSubNodesToFile  (ParentNode : TTreeNode; FileName : String) : Boolean;

    {Save and restore the subnodes expanded state:}
    Procedure SaveNodesState(Node : TTreeNode);
    Procedure RestoreNodesState(Node : TTreeNode);

    {Menu-handling:}
    Procedure DisplayContextMenu(Node : TTreeNode);       Dynamic;
    Procedure DisplayPropertiesMenu(Node : TTreeNode);    Dynamic;

    {Watchthread handling:}
    Procedure StartWatchThread;                           Virtual;
    Procedure StopWatchThread;                            Virtual;
    Procedure TerminateWatchThread(Drive : TDrive);       Virtual;
    Procedure StartAllWatchThreads;                       Virtual;
    Procedure StopAllWatchThreads;                        Virtual;
    Function  WatchThreadActive : Boolean;                Virtual;
    Function  NodeWatched(Node : TTreeNode) : Boolean;    Virtual;


    (* Modified Events: *)
    Procedure GetImageIndex(Node: TTreeNode);             Override;
    Function  CanEdit(Node: TTreeNode) : Boolean;         Override;
    Function  CanChange(Node: TTreeNode): Boolean;        Override;
    Function  CanExpand(Node: TTreeNode): Boolean;        Override;
    Procedure Delete(Node: TTreeNode);                    Override;
    Procedure Loaded;                                     Override;
    Procedure KeyDown(var Key: Word; Shift: TShiftState); Override;
    Procedure Change(Node: TTreeNode);                    Override;
    Procedure KeyPress(Var Key : Char);                   Override;

{---------------------------------------------------------------}
  published
{---------------------------------------------------------------}
   {Additional properties:}

   {Current selected directory}
   Property Directory   : String          Read GetDirectory
                                          Write SetDirectory;
   {Confirm deleting directories:}
   Property ConfirmDelete       : Boolean Read fConfirmDelete
                                          Write fConfirmDelete
                                          Default True;

   {Confirm overwriting directories:}
   Property ConfirmOverwrite    : Boolean Read fConfirmOverwrite
                                          Write fConfirmOverwrite
                                          Default True;

   {Scan all directories in method ScanDrive:}
   Property FullDriveScan     : Boolean Read fFullDriveScan
                                        Write SetFullDriveScan;

   Property DimmHiddenDirs     : Boolean Read fDimmHiddenDirs
                                        Write SetDimmHiddenDirs;

   {Enable automatic update on filesystem changes:}
   Property WatchDirectory    : Boolean Read fWatchDirectory
                                        Write SetWatchDirectory;

   {Peform automatic update after ChangeInterval milliseconds:}
   Property ChangeInterval    : Cardinal Read fChangeInterval
                                         Write SetChangeInterval
                                         Default 1000;

   {Enables or disables the system context menu for a directory:}
   Property UseSystemContextMenu : Boolean Read FUseSystemContextMenu
                                           Write FUseSystemContextMenu
                                           Default True;

   {Linked component TDirView:}
   Property DirView           : TDirView Read fDirView
                                         Write SetDirView;

   Property ColorBold         : TColor  Read fColorBold
                                        Write fColorBold
                                        Default clBlue;

   Property ShowDirSize       : Boolean Read fShowDirSize
                                        Write SetShowDirSize;

   {Show the volume labels of drives:}
   Property ShowVolLabel      : Boolean Read fShowVolLabel
                                        Write SetShowVolLabel;

   {How to display the drives volume labels:}
   Property VolDisplayStyle     : TVolumeDisplayStyle Read fVolDisplayStyle
                                                      Write SetVolDisplayStyle
                                                      Default doPrettyName;

   {Show AVI-animation when performing a full drive scan:}
   Property ShowAnimation    : Boolean Read  FShowAnimation
                                       Write FShowAnimation;

   {Additional events:}
   Property OnStartScan : TNotifyEvent Read fOnStartScan
                                       Write fOnStartScan;

   Property OnEndScan   : TNotifyEvent Read fOnEndScan
                                       Write fOnEndScan;

   Property OnScanDir   : TDriveViewScanDirEvent    Read fOnScanDir
                                                    Write fOnScanDir;
   Property OnDiskChange: TDriveViewDiskChangeEvent Read fOnDiskChange
                                                    Write fOnDiskChange;

   Property OnChangeDetected : TDriveViewDiskChangeEvent Read fOnChangeDetected
                                                         Write fOnChangeDetected;

   Property OnChangeInvalid  : TDriveViewDiskChangeEvent Read fOnChangeInvalid
                                                         Write fOnChangeInvalid;

{$IFDEF DRAGDROP}
    {Drag&Drop properties:}
    {Show drag images during a drag&drop operation:}
    Property UseDragImages         : Boolean      Read FUseDragImages
                                                  Write FUseDragImages
                                                  Default True;

    {Show popupmenu when dropping a file with the right mouse button:}
    Property TargetPopUpMenu       : Boolean      Read FTargetPopUpMenu
                                                  Write SetTargetPopUpMenu
                                                  Default True;

    {The mouse has entered the component window as a target of a drag&drop operation:}
    Property OnDDDragEnter         : TDDOnDragEnter Read FOnDDDragEnter
                                                    Write FOnDDDragEnter;

    {The mouse has leaved the component window as a target of a drag&drop operation:}
    Property OnDDDragLeave         : TDDOnDragLeave Read FOnDDDragLeave
                                                    Write FOnDDDragLeave;

    {The mouse is dragging in the component window as a target of a drag&drop operation:}
    Property OnDDDragOver          : TDDOnDragOver  Read FOnDDDragOver
                                                    Write FOnDDDragOver;
    {The Drag&drop operation is about to be executed:}
    Property OnDDDrop              : TDDOnDrop      Read FOnDDDrop
                                                    Write FOnDDDrop;

    Property OnDDQueryContinueDrag : TDDOnQueryContinueDrag  Read FOnDDQueryContinueDrag
                                                             Write FOnDDQueryContinueDrag;

    Property OnDDGiveFeedback      : TDDOnGiveFeedback Read FOnDDGiveFeedback
                                                       Write FOnDDGiveFeedback;

    {A drag&drop operation is about to be initiated whith the components window as the
     source:}
    Property OnDDDragDetect        : TDDOnDragDetect   Read FOnDDDragDetect
                                                       Write FOnDDDragDetect;

    {The component window is the target of a drag&drop operation:}
    Property OnDDProcessDropped    : TOnProcessDropped Read FOnDDProcessDropped
                                                       Write FOnDDProcessDropped;

    {An error has occured during a drag&drop operation:}
    Property OnDDError             : TDDErrorEvent Read FOnDDError
                                                   Write FOnDDError;

    {The drag&drop operation has been executed:}
    Property OnDDExecuted          : TDDExecutedEvent Read FOnDDExecuted
                                                      Write FOnDDExecuted;
    {Event is fired just before executing the fileoperation. This event is also fired when
     files are pasted from the clipboard:}
    Property OnDDFileOperation     : TDDFileOperationEvent Read FOnDDFileOperation
                                                           Write FOnDDFileOperation;

    {Event is fired after executing the fileoperation. This event is also fired when
     files are pasted from the clipboard:}
    Property OnDDFileOperationExecuted  : TDDFileOperationExecutedEvent Read FOnDDFileOperationExecuted
                                                                        Write FOnDDFileOperationExecuted;
{$ENDIF}

   property Align;
   property Anchors;
   property AutoExpand;
   property BiDiMode;
   property BorderStyle;
   property BorderWidth;
   property ChangeDelay;
   property Color;
   property Ctl3D;
   property Constraints;
   {Delphi's drag&drop is not compatible with the OLE windows drag&drop:}
{$IFNDEF DRAGDROP}
   property DragKind;
   property DragCursor;
   property DragMode;
   property OnDragDrop;
   property OnDragOver;
{$ENDIF}
   property Enabled;
   property Font;
   property HideSelection;
   property HotTrack;
   property Indent;
   property ParentBiDiMode;
   property ParentColor;
   property ParentCtl3D;
   property ParentFont;
   property ParentShowHint;
   property PopupMenu;
   property ReadOnly;
   property RightClickSelect;
   property RowSelect;
   property ShowButtons;
   property ShowHint;
   property ShowLines;
  {property ShowRoot;}
  {property SortType;}
   property TabOrder;
   property TabStop;
   property ToolTips;
   property Visible;
   property OnChange;
   property OnChanging;
   property OnClick;
   property OnCollapsing;
   property OnCollapsed;
   property OnCompare;
   {Internal used events:
   property OnCustomDraw;
   property OnCustomDrawItem;}
   property OnDblClick;
   property OnDeletion;
   property OnEdited;
   property OnEditing;
   property OnEndDock;
   property OnEndDrag;
   property OnEnter;
   property OnExit;
   property OnExpanding;
   property OnExpanded;
   property OnGetImageIndex;
   property OnGetSelectedIndex;
   property OnKeyDown;
   property OnKeyPress;
   property OnKeyUp;
   property OnMouseDown;
   property OnMouseMove;
   property OnMouseUp;
   property OnStartDock;
   property OnStartDrag;
  end;
{---------------------------------------------------------------}



// ===========================================================
// Other service procedures and functions:}
// ===========================================================

procedure Register;



{==============================================================}
implementation
{==============================================================}
Uses IEComboBox;


ResourceString
   English_ErrorInvalidDirName = 'New name contains Invalid characters:';
   German_ErrorInvalidDirName  = 'Verzeichnisname enthlt ungltige Zeichen:';
   French_ErrorInvalidDirName = 'Le nouveau nom contient des caractres invalides:';

Type
  PInt      = ^Integer;

  TLogFileNode = Record
       Level : Integer;
       Attrs : Integer;
       ShortName : array[0..13] of AnsiChar;
       NameLen : Integer;
  End;

  TLogFileHeader = Record
       ID : String[10];
       Version : String[3];
  End;

// ===========================================================
// Global variables
// ===========================================================

Const ErrorNodeNA = '%s: Node not assigned';

Var LogFileNode         : TLogFileNode;
    TempFile            : File;
    ErrorInvalidDirName : String;




procedure Register;
begin
  RegisterComponents('IE', [TDriveView]);
end;  {Register}


Constructor TNodeData.Create;
Begin
  Inherited Create;
  fAttr         := 0;
  fDrawBold     := False;
  fExpanded     := False;
  fNewDir       := False;
  fDirName      := '';
  fShortName    := '';
  fDirSize      := InvalidSize;
  fIsRecycleBin := False;
End; {TNodeData.Create}


Destructor TNodeData.Destroy;
Begin
  SetLength(fDirName, 0);
  Inherited Destroy;
End; {TNodeData.Destroy}



// ===========================================================
// Class TDriveView:
// ===========================================================


(* -------------------------*)
(* Events:                  *)
(* -------------------------*)

(* Overwrite Event OnCustomDraw: *)


Procedure TDriveview.InternalOnDrawItem(Sender: TCustomTreeView; Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
Begin
  IF Assigned(Node.Data) And (Node <> DropTarget) Then
  Begin
    {Dimmed display, if hidden-atrribut set:}
    IF fDimmHiddenDirs And Not Node.Selected And
      (TNodeData(Node.Data).Attr And faHidden <> 0) Then
    Canvas.Font.Color := clGrayText
    Else
    {HideSelection:}
    IF Not Self.Focused And Node.Selected And HideSelection Then
    Begin
        Canvas.Brush.Color := clBtnFace;
        Canvas.Font.Color := clBtnText;
    End;

    IF TNodeData(Node.Data).DrawBold Then
    Begin
      Canvas.Font.Color := fColorBold;
      Canvas.Font.Style := Canvas.Font.Style + [fsBold];
    End;
  End;
End; {InternalOnDrawItem}


(* Overwrite Event OnEditing: *)

Function TDriveView.CanEdit(Node: TTreeNode) : Boolean;
Begin
  Result := Inherited CanEdit(Node);
  IF Result Then
  Result := Assigned(Node.Parent) And
            Not TNodeData(Node.Data).isRecycleBin And
            Not ReadOnly And
{$IFDEF DRAGDROP}
            (FDD.DragDetectStatus <> ddsDrag) And
{$ENDIF}
            (TNodeData(Node.Data).Attr and (faReadOnly or faSysFile) = 0) And
            (GetDriveTypetoNode(Node) <> DRIVE_CDROM) And
            (Node.Text = GetDirName(Node));
End; {CanEdit}


(* Overwrite Event OnCanChange: *)

Function  TDriveView.CanChange(Node: TTreeNode): Boolean;
Begin
  Result := Inherited CanChange(Node);
  IF Result              And
     Not FCanChange      And
     Assigned(Node)      And
     Assigned(Node.Data) And
     Assigned(Selected)  And
     Assigned(Selected.Data) Then
  Begin
    DropTarget := Node;
    Result := False;
  End
  Else
  DropTarget := NIL;
End; {CanChange}


(* Overwrite Event OnExpanding: *)

Function  TDriveView.CanExpand(Node: TTreeNode): Boolean;
Var SubNode : TTreeNode;
    Drive   : TDrive;
Begin
  Result := Inherited CanExpand(Node);
  Drive := GetDrive(Node);
  IF Node.HasChildren Then
  Begin
    IF (Node.Level = 0) And
        Not DriveStatus[Drive].Scanned  And
        (Drive >= FirstFixedDrive)      Then
    Begin
      SubNode := Node.GetFirstChild;
      IF Not Assigned(SubNode) Then
      Begin
        ScanDrive(Drive);
        SubNode             := Node.GetFirstChild;
        Node.HasChildren    := Assigned(SubNode);
        Result              := Node.HasChildren;
        IF Not Assigned(DriveStatus[Drive].DiscMonitor) Then
        CreateWatchThread(Drive);
      End;
    End
    Else
    Begin
      SubNode             := Node.GetFirstChild;
      Node.HasChildren    := Assigned(SubNode);
      ValidateDirectoryEasy(Node);
    End;
  End;
End; {CanExpand}


(* Overwrite event OnGetImageIndex: *)

procedure TDriveView.GetImageIndex(Node: TTreeNode);
Begin
  IF Node.ImageIndex = 0 Then
  SetImageIndex(Node);
  Inherited GetImageIndex(Node);
End; {GetImageIndex}


(* Overwrite event Loaded: *)

Procedure TDriveView.Loaded;
Begin
  Inherited Loaded;
  {Create the drive nodes:}
  RefreshRootNodes(False, dsDisplayName or dvdsFloppy);
  {Set the initial directory:}
  IF Length(FDirectory) > 0 Then
  Directory := FDirectory;

  fCreating := FALSE;
{$IFDEF DRAGDROP}
  FDD.DragDropControl := Self;
{$ENDIF}
End; {Loaded}


(* event OnEdited: *)

procedure TDriveView.Edit(const Item: TTVItem);
Var NewDirName : String;
    SRec       : TSearchRec;
    Node       : TTreeNode;
    Info       : String;
    i          : Integer;

Begin
  Node := GetNodeFromHItem(Item);
  IF (Length(Item.pszText) > 0) And (Item.pszText <> Node.Text) Then
  Begin
    IF StrContains(coInvalidDosChars, Item.pszText) Then
    Begin
      Info := coInvalidDosChars;
      For i := Length(Info) DownTo 1 Do
      System.Insert(Space, Info, i);

      IF Length(Item.pszText) > 0 Then
        Raise EInvalidDirName.Create(ErrorInvalidDirName + Space + Info);
      Exit;
    End;

    StopWatchThread;

    IF RenameFile(GetDirPath(Node), AddSlash(GetDirPath(Node.Parent)) + Item.pszText) Then
    Begin
      Node.Text := Item.pszText;
      TNodeData(Node.Data).Dirname := Item.pszText;
      IF FindFirst(NewDirName, faAnyFile, Srec) = 0 Then
        TNodeData(Node.Data).ShortName := Srec.FindData.cAlternateFileName;
      FindClose(Srec);
      SortChildren(Node.Parent, False);
      inherited Edit(Item);
      IF Assigned(DirView) Then
      DirView.Path := GetDirPathName(Selected);
    End
    Else
    Raise ERenameFileFailed.Create(ErrorRenameFile + #10 + Item.pszText);
    StartWatchThread;
  End;
End; {Edit}


(* Overwrite event OnDeletion: *)
Procedure TDriveView.Delete(Node: TTreeNode);
Var NodeData : TNodeData;

Begin
{$IFDEF DRAGDROP}
  If Node = DragNode Then
  DragNode := NIL;
{$ENDIF}
  IF Node = DropTarget Then
  Begin
    DropTarget := NIL;
    Application.ProcessMessages;
  End;

  NodeData := NIL;

  IF Assigned(Node) And Assigned(Node.Data) Then
  NodeData := TNodeData(Node.Data);
  Node.Data := NIL;

  Inherited Delete(Node);

  If Assigned(NodeData) Then
  NodeData.Destroy;
End; {OnDelete}


(* Overwrite event OnKeyDown: *)

procedure TDriveView.KeyDown(var Key: Word; Shift: TShiftState);
Begin
  Inherited KeyDown(Key, Shift);
  IF (Key = VK_RETURN) And
     (ssAlt in Shift)  And
     Not isEditing     And
     Assigned(Selected) Then
     DisplayPropertiesMenu(Selected);
End; {KeyDown}


(* Overwrite event OnKeyPress: *)

Procedure TDriveView.KeyPress(Var Key : Char);
Begin
  Inherited KeyPress(Key);
  IF Assigned(Selected) Then
  Case Key of
    #13: Begin
           Selected.Expanded := Not Selected.Expanded;
           Key := #0;
           Exit;
         End;
    '/': Begin
           Selected.Collapse(True);
           Selected.MakeVisible;
           Key := #0;
           Exit;
         End;
    '*': Selected.MakeVisible;

    End; {Case}
End; {KeyPress}



(* Overwrite event OnChange: *)

Procedure TDriveView.Change(Node: TTreeNode);
Var Drive     : TDrive;
    OldSerial : DWORD;
    NewDir    : String;
    LastDrive : TDrive;

Begin
  IF Assigned(Node) Then
  Begin
    NewDir := GetDirPath(Node);
    IF NewDir <> FLastDir Then
    Begin
      Drive := NewDir[1];
      IF Length(FLastDir) > 0 Then
      LastDrive := FLastDir[1]
      Else
      LastDrive := #0;

      OldSerial := DriveInfo[Drive].DriveSerial;
      DriveInfo.ReadDriveStatus(Drive, dsSize or dsImageIndex);
      With DriveInfo[Drive] Do
      Begin
        IF Assigned(fDirView) And (fDirView.Path <> NewDir) Then
          fDirView.Path := NewDir;
        IF Assigned(FDriveBox) Then
          TIEDriveComboBox(FDriveBox).Drive := GetDrive(Selected);
        IF DriveReady Then
        Begin
          IF LastDrive <> Drive Then
          Begin
            IF (LastDrive = 'A') Or (LastDrive = 'B') Then
              TerminateWatchThread(LastDrive);
            IF (DriveSerial <> OldSerial) Then
            ScanDrive(Drive);
            If Assigned(fOnDiskChange) Then
            fOnDiskChange(Self, Drive);
            StartWatchThread;
          End
          Else
          Begin
            IF Not DirExists(NewDir) Then
            Begin
              ValidateDirectory(DriveStatus[Drive].RootNode);
              IF Directory <> NewDir Then
              Exit;
            End;
            StartWatchThread;
          End;
        End
        Else
        Begin
          DriveStatus[Drive].RootNode.DeleteChildren;
          If (LastDrive <> Drive) And Assigned(fOnDiskChange) Then
          fOnDiskChange(Self, Drive);
        End;
      End;
      fChangeFlag := True;
      fLastDir := NewDir;
    End;
  End;
  Inherited Change(Node);
End; {Change}


// ===========================================================
// Methods of object TDriveView:
// ===========================================================

constructor TDriveView.Create(AOwner: TComponent);
Var Drive : TDrive;
    WinVer : TOSVersionInfo;


Begin
  Inherited Create(AOwner);
  fCreating := TRUE;
  WinVer.dwOSVersionInfoSize := SizeOf(WinVer);
  GetVersionEx(WinVer);

  IF fChangeInterval = 0 Then
  fChangeInterval       := 1000;

  For Drive := FirstDrive To LastDrive Do
  With DriveStatus[Drive] Do
  Begin
    Scanned      := False;
    Verified     := False;
    RootNode     := NIL;
    DiscMonitor  := NIL;
    {ChangeTimer: }
    ChangeTimer := TTimer.Create(Self);
    ChangeTimer.Interval := 0;
    ChangeTimer.Enabled  := False;
    ChangeTimer.OnTimer  := ChangeTimerOnTimer;
    ChangeTimer.Tag      := Ord(Drive);
  End;

  FFileOperator := TFileOperator.Create(Self);
  FFileOperator.ProgressTitle := coFileOperatorTitle;
  FFileOperator.Flags := [foAllowUndo, foNoConfirmMkDir];


  FColorBold         := clBlue;
  FShowVolLabel      := True;
  FChangeFlag        := False;
  FContextMenu       := False;
  FLastDir           := EmptyStr;
  FValidateFlag      := False;
  FConfirmDelete     := True;
  FUseSystemContextMenu := True;
  FCanChange         := True;
  FContinue          := True;
  FShowAnimation     := False;

{$IFDEF DRAGDROP}
  {Drag&drop:}
  DragMode          := dmAutomatic;
  fConfirmOverwrite := True;
  FDragDrive        := #0;
  DragFileList      := TStringList.Create;
  FLastPathCut      := '';
  FTargetPopupMenu  := True;
  FUseDragImages    := (Win32PlatForm = VER_PLATFORM_WIN32_NT) Or (WinVer.dwMinorVersion > 0);
  FStartPos.X       := -1;
  FStartPos.Y       := -1;
  FDragPos          := FStartPos;
  FExeDrag          := False;

  FDD := TDragDropFilesEx.Create(Self);
  With FDD Do
  Begin
    AcceptOwnDnd     := True;
    AutoDetectDnD    := False;
    BringToFront     := True;
    CompleteFileList := True;
    NeedValid        := [nvFileName];
    RenderDataOn     := rdoEnterAndDropSync;
    SourceEffects    := [deCopy, deMove, deLink];
    TargetEffects    := [deCopy, deMove, deLink];
    TargetPopUpMenu  := FTargetPopupMenu;
    OnDragDetect     := DDDragDetect;
    OnDragEnter      := DDDragEnter;
    OnDragLeave      := DDDragLeave;
    OnDragOver       := DDDragOver;
    OnProcessDropped := DDProcessDropped;
    OnDrop           := DDDrop;
    OnQueryContinueDrag := DDQueryContinueDrag;
    OnGiveFeedback   := DDGiveFeedback;
  End;
{$ENDIF}
  OnCustomDrawItem      := InternalOnDrawItem;
End; {Create}


Procedure TDriveView.CreateWnd;
Var FileInfo        : TShFileInfo;

Begin
  Inherited CreateWnd;
  IF Not Assigned(Images) Then
  Begin
    Images := TImageList.Create(Self);
    Images.Handle := SHGetFileInfo('', 0, FileInfo, SizeOf(FileInfo), SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
    Images.ShareImages := True;
  End;

  IF Not Assigned(StateImages) Then
  Begin
    StateImages := TImageList.Create(Self);
    StateImages.Handle := SHGetFileInfo('', 0, FileInfo, SizeOf(FileInfo), SHGFI_SYSICONINDEX or SHGFI_OPENICON);
    StateImages.ShareImages := True;
  End;

  IF Assigned(PopupMenu) Then
  PopupMenu.Autopopup := False;
  FParentForm := GetParentForm(Self);
  IF Not Assigned(DragImageList) And Assigned(FParentForm) And Not (csDesigning in ComponentState) Then
  DragImageList := TImageList.Create(FParentForm);
End; {CreateWnd}


Destructor TDriveView.Destroy;
Var Drive : TDrive;
Begin
  IF Assigned(DragImageList) Then
  DragImageList.Free;
  DragImageList := NIL;

  IF Assigned(Images) Then
  Images.Free;

  IF Assigned(StateImages) Then
  StateImages.Free;

  For Drive := FirstDrive To LastDrive Do
  With DriveStatus[Drive] Do
  Begin
    IF Assigned(DiscMonitor) Then
       Discmonitor.Free;
    IF Assigned(ChangeTimer) Then
       ChangeTimer.Free;
  End;

  IF Assigned(FFileOperator) Then
     FFileOperator.Free;

{$IFDEF DRAGDROP}
  DragFileList.Destroy;

  IF Assigned(FDD) Then
  FDD.Free;
{$ENDIF}
  Inherited Destroy;
End; {Destroy}


Function TDriveView.GetNodeFromHItem(Item: TTVItem): TTreeNode;
begin
  with Item do
    if (state and TVIF_PARAM) <> 0 then
      Result := Pointer(lParam)
    else
      Result := Items.GetNode(hItem);
end; {GetNodeFromItem}


{$IFDEF DRAGDROP}
Procedure TDriveView.CNNotify(Var Msg: TWMNotify);
Begin
  Case Msg.NMHdr.code Of
    TVN_BEGINDRAG:  DDDragDetect(MK_LBUTTON, FStartPos, Mouse.CursorPos, ddsDrag);
    TVN_BEGINRDRAG: DDDragDetect(MK_RBUTTON, FStartPos, Mouse.CursorPos, ddsDrag);
    Else
    Inherited;
  End;
End; {CNNotify}
{$ENDIF}


Procedure TDriveView.WMLButtonDown(var Msg: TWMLButtonDown);
Begin
  FCanChange := False;
{$IFDEF DRAGDROP}
  GetCursorPos(FStartPos);
{$ENDIF}
  Inherited;
End; {WMLButtonDown}


Procedure TDriveView.WMLButtonUp(var Msg: TWMLButtonDown);
Begin
  FCanChange := True;
  IF Assigned(DropTarget) And Assigned(DropTarget.Data) Then
    Selected   := DropTarget;
  DropTarget := NIL;
  Inherited;
End; {WMLButtonUp}


Procedure TDriveView.WMRButtonDown(var Msg: TWMRButtonDown);
Begin
{$IFDEF DRAGDROP}
  GetCursorPos(FStartPos);
  IF FDD.DragDetectStatus <> ddsDrag Then
{$ENDIF}
  fContextMenu := True;
  Inherited;
End; {WMRButtonDown}


Procedure TDriveView.WMRButtonUp(Var Msg: TWMRButtonUp);
Var Node       : TTreeNode;
    DirWatched : Boolean;
    P          : TPoint;

Begin
  IF Assigned(PopupMenu) Then
  PopupMenu.Autopopup := False;
{$IFDEF DRAGDROP}
  FStartPos.X := -1;
  FStartPos.Y := -1;
{$ENDIF}

  IF fContextMenu Then
  Begin
    Node:=GetNodeAt(Msg.XPos, Msg.YPos);
{$IFDEF DRAGDROP}
    FDD.AutoDetectDnD := False;
{$ENDIF}
    IF FUseSystemContextMenu And Assigned(Node) Then
    Begin
      DirWatched := NodeWatched(Node) And WatchThreadActive;
      DisplayContextMenu(Node);
      IF Not DirWatched Then
      ValidateDirectory(Node);
    End
    Else
    Begin
      P.X := Msg.XPos;
      P.Y := Msg.YPos;
      P := ClientToScreen(P);
      IF Assigned(PopupMenu) And Not PopupMenu.AutoPopup Then
        PopupMenu.Popup(P.X, P.Y);
    End;
{$IFDEF DRAGDROP}
    FDD.AutoDetectDnD := True;
{$ENDIF}
  End;
  fContextMenu := False;
  Inherited;
End; {WMRButtonUp}



Procedure TDriveView.SetImageIndex(Node : TTreeNode);
Var FileInfo  : TShFileInfo;
    NodePath  : String;

Begin
  IF Assigned(Node) And (Node.ImageIndex = 0) Then
  Begin
    NodePath := GetDirPathName(Node);
    IF Node.Level = 0 Then
    Begin
      With DriveInfo[NodePath[1]] Do
      Begin
        IF ImageIndex = 0 Then
        Begin
          DriveInfo.ReadDriveStatus(NodePath[1], dsImageIndex);
          Node.ImageIndex    := DriveInfo[NodePath[1]].ImageIndex;
        End
        Else
        Node.ImageIndex    := ImageIndex;
        Node.SelectedIndex := Node.ImageIndex;
      End;
    End
    Else
    Begin
      IF (DriveInfo[NodePath[1]].DriveType = DRIVE_REMOTE) Then
      Begin
        Node.ImageIndex    := StdDirIcon;
        Node.SelectedIndex := StdDirSelIcon;
      End
      Else
      Begin
        Try
          SHGetFileInfo(PChar(NodePath), 0, FileInfo, SizeOf(FileInfo),
                         SHGFI_SYSICONINDEX or SHGFI_SMALLICON);

          IF (FileInfo.iIcon < Images.Count) And (FileInfo.iIcon > 0) Then
          Begin
            Node.ImageIndex    := FileInfo.iIcon;
            SHGetFileInfo(PChar(NodePath), 0, FileInfo, SizeOf(FileInfo),
                           SHGFI_SYSICONINDEX or SHGFI_SMALLICON OR SHGFI_OPENICON );
            Node.SelectedIndex    := FileInfo.iIcon;
          End
          Else
          Begin
            Node.ImageIndex    := StdDirIcon;
            Node.SelectedIndex := StdDirSelIcon;
          End;
        Except Else
          Begin
            Node.ImageIndex    := StdDirIcon;
            Node.SelectedIndex := StdDirSelIcon;
          End;
        End;
      End;
    End;
  End; {ImageIndex = 0}
End; {SetImageIndex}


Function TDriveView.GetDriveText(Drive : TDrive) : String;
Begin
  With DriveInfo[Drive] Do
  Begin
    IF fShowVolLabel And (Length(PrettyName) >  0) Then
    Begin
      Case fVolDisplayStyle Of
      doPrettyName:     Result := Prettyname;
      doDisplayName:    Result := DisplayName;
      doLongPrettyName: Result := LongPrettyname;
      End; {Case}
    End Else
      Result := Drive + ':';
  End;
End; {GetDriveText}


Function TDriveView.GetValidDrivesStr : String;
Var Drive : TDrive;

Begin
  Result := '';
  For Drive := FirstDrive to LastDrive Do
  IF DriveInfo[Drive].Valid Then
  Result := Result + Drive;
End; {GetValidDriveStr}



Procedure TDriveView.RefreshRootNodes(ScanDirectory : Boolean; dsFlags : Integer);
Var Drive   : Char;
    NewText : String;
    NextDrive : TDrive;
    D         : TDrive;
    SaveCursor: TCursor;
    WasValid  : Boolean;
    OldSerial : DWORD;
    wFirstDrive : TDrive;
    NodeData    : TNodeData;

Begin
  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  Try
    IF dsFlags And dvdsFloppy <> 0 Then
    wFirstDrive := FirstDrive
    Else
    wFirstDrive := FirstFixedDrive;
    For Drive := wFirstDrive to LastDrive Do
    Begin
      With DriveInfo[Drive] Do
      Begin
        WasValid  := Valid And Assigned(FindNodeToPath(Drive + ':\'));
        OldSerial := DriveSerial;
      End;
      IF (dsFlags And dvdsReReadAllways = 0) And (Length(DriveInfo[Drive].DisplayName) > 0) Then
      dsFlags := dsFlags And Not dsDisplayName;

      DriveInfo.ReadDriveStatus(Drive, dsFlags);
      With DriveInfo[Drive], DriveStatus[Drive] Do
      Begin
        IF Valid Then
        Begin
          IF Not WasValid Then
          {New drive has arrived: insert new rootnode:}
          Begin
            NextDrive := LastDrive;
            IF Not fCreating Then
            For D := Drive To LastDrive Do
            Begin
              IF Assigned(DriveStatus[D].RootNode) Then
              Begin
                NextDrive := D;
                Break;
              End;
            End;

            { Create root directory node }
            NodeData := TNodeData.Create;
            NodeData.DirName   := Drive + ':\';
            NodeData.ShortName := Drive + ':\';

            IF Assigned(DriveStatus[NextDrive].RootNode) Then
            RootNode := Items.InsertObject(DriveStatus[NextDrive].RootNode, '', NodeData)
            Else
            RootNode := Items.AddObject(nil, '', NodeData);

            RootNode.Text        := GetDisplayName(RootNode);
            RootNode.HasChildren := TRUE;

            Scanned              := False;
            Verified             := False;
          End
          Else
          If RootNode.ImageIndex <> DriveInfo[Drive].ImageIndex Then
          Begin
            RootNode.ImageIndex    := DriveInfo[Drive].ImageIndex;
            RootNode.SelectedIndex := DriveInfo[Drive].ImageIndex;
          End;


          IF Drive >= FirstFixedDrive Then
          Begin
            IF ScanDirectory And (DriveSerial <> OldSerial) Then
            ScanDrive(Drive);
          End;

          IF Assigned(RootNode) Then
          Begin
            NewText := GetDisplayName(RootNode);
            IF RootNode.Text <> NewText Then
            RootNode.Text := NewText;
          End;
        End
        Else
        IF WasValid And Not Valid Then
        {Drive has been removed => delete rootnode:}
        Begin
          IF Directory[1] = Drive Then
          Begin
            Directory := GetDirPathName(DriveStatus[Drive].RootNode.GetPrevSibling);
            IF Not Assigned(Selected) Then
            Directory := GetDirPathName(DriveStatus[FirstFixedDrive].RootNode);
          End;
          Scanned     := False;
          Verified    := False;
          RootNode.Delete;
          RootNode := NIL;
        End;
      End;
    End;
  Finally
    Screen.Cursor := SaveCursor;
  End;
End; {RefreshRootNodes}


Function TDriveView.AddChildNode(ParentNode : TTreeNode; Srec : TSearchRec) : TTreeNode;
Var NewNode  : TTreeNode;
    NodeData : TNodeData;

Begin
  NodeData := TNodeData.Create;
  NodeData.Attr          := Srec.Attr;
  NodeData.DirName       := Srec.Name;
  NodeData.ShortName     := Srec.FindData.cAlternateFileName;
  NodeData.fisRecycleBin := (Srec.Attr And faSysFile <> 0) And
                                         (ParentNode.Level = 0) And
                                         ((UpperCase(Srec.Name) = 'RECYCLED') Or (UpperCase(Srec.Name) = 'RECYCLER'));
  NewNode := Self.Items.AddChildObject(ParentNode, '', NodeData);
  NewNode.Text := GetDisplayName(NewNode);
  Result := NewNode;
End; {AddChildNode}



Function  TDriveView.GetDriveStatus(Drive : TDrive) : TDriveStatus;
Begin
  Result := DriveStatus[Upcase(Drive)];
End; {GetDriveStatus}



Function TDriveView.DoScanDir(FromNode : TTreeNode) : Boolean;
Var ScanThisDir : Boolean;
Begin
  With TNodeData(FromNode.Data) Do
  ScanThisDir := Not isRecycleBin;
  IF Assigned(fOnScanDir) Then
  fOnScanDir(Self, FromNode, ScanThisDir);
  Result := ScanThisDir;
End; {DoScanDir}


Procedure TDriveView.ScanDrive(Drive : TDrive);
Var DosError   : Integer;
    RootNode   : TTreeNode;
    SaveCursor : TCursor;
    FAnimate   : TAnimate;

Procedure ScanPath(Const Path : String; ParentNode : TTreeNode);
Var Srec      : TSearchRec;
    SubNode   : TTreeNode;

Begin
  IF Not DoScanDir(ParentNode) Then
  Exit;

  DosError := FindFirst(Path, faDirectory or faHidden or FaSysFile, Srec);
  While DosError = 0 Do
  Begin
     IF (Srec.Name <> '.') And
        (Srec.Name <> '..') And
        (Srec.Attr And faDirectory <> 0) Then
     Begin
       IF (Srec.Attr And faDirectory) <> 0 Then
       Begin { Scan subdirectory }
         SubNode := AddChildNode(ParentNode , Srec);
         TNodeData(SubNode.Data).Verified := True;
         ScanPath(ExtractFilePath(Path) + Srec.Name + '\*.*', SubNode );
         IF Not FContinue Then
         Break;
       End;
     End;
     DosError := FindNext(Srec);
  End;
  FindClose(Srec);
  IF (Items.Count Mod 10) = 0 Then
  Application.ProcessMessages;
  IF Not FContinue Then
  Exit;
End; {ScanPath}


Begin {ScanDrive}
  with Self.Items do
  begin
    FContinue := True;
    IF Not fFullDriveScan Then
    Begin
      ValidateDirectory(FindNodeToPath(Drive + ':\'));
      DriveStatus[Drive].Scanned  := TRUE;
      DriveStatus[Drive].Verified := FALSE;
    End
    Else
    Begin
      FAnimate      := NIL;
      SaveCursor    := Screen.Cursor;
      Screen.Cursor := crHourglass;
      Items.BeginUpdate;

      IF FShowAnimation Then
      Begin
        FAnimate := TAnimate.Create(Self);
        FAnimate.Top  := (Height - FAnimate.Height) DIV 2;
        FAnimate.Left := ((Width - FAnimate.Width) * 2) DIV 3;
        FAnimate.Parent := Self;
        FAnimate.CommonAVI := aviFindFolder;
        FAnimate.Active := True;
      End;

      If Assigned(fOnStartScan) Then
      fOnStartScan(Self);
      Try
        RootNode := DriveStatus[Drive].RootNode;
        IF Not Assigned(RootNode) Then Exit;

        IF RootNode.HasChildren Then
          RootNode.DeleteChildren;

        ScanPath(Drive + ':\*.*', RootNode);      { scan subdirectories of rootdir}
        TNodeData(RootNode.Data).Verified := True;

        DriveStatus[Drive].Scanned  := TRUE;
        DriveStatus[Drive].Verified := TRUE;
      finally
        AlphaSort;
        EndUpdate;
        IF Assigned(FAnimate) Then
        FAnimate.Free;
      End;
      RootNode.Expand(False);

      Screen.Cursor := SaveCursor;
      If Assigned(FOnEndScan) Then
      FOnEndScan(Self);
    End;
  End;
End; {ScanDrive}


Function TDriveView.FindNodeToPath(Path : String) : TTreeNode;
Var Drive: Char;

Function SearchSubDirs(ParentNode : TTreeNode; Path : String) : TTreeNode;
Var i    : Integer;
    Node : TTreeNode;
    Dir  : String;

Begin
  Result := NIL;
  IF Length(Path) = 0 Then
  Exit;

  {Extract first directory from path:}
  i := Pos('\', Path);

  IF (i = 0) Then
  i := Length(Path);

  Dir := System.Copy(Path, 1, i);
  System.Delete(Path, 1, i);

  IF Dir[Length(Dir)] = '\' Then
  SetLength(Dir, Pred(Length(Dir)));


  Result := NIL;
  Node := ParentNode.GetFirstChild;
  IF Not Assigned(Node) Then
  Begin
    ValidateDirectory(ParentNode);
    Node := ParentNode.GetFirstChild;
  End;

  While Assigned(Node) Do
  Begin
    IF (UpperCase(GetDirName(Node)) = Dir) OR (TNodeData(Node.Data).ShortName = Dir) Then
    Begin
      IF Length(Path) > 0 Then
      Result := SearchSubDirs(Node, Path)
      Else
      Result := Node;
      Exit;
    End;
    Node := ParentNode.GetNextChild(Node);
  End;

End; {SearchSubDirs}


Begin {FindNodeToPath}
  Result := NIL;
  IF Length(Path) < 3 Then
  Exit;
  Drive := UpCase(Path[1]);
  IF (Drive < FirstDrive) Or (Drive > LastDrive) Then
  EConvertError.Create(Format(ErrorInvalidDrive, [Drive]));

  IF Assigned(DriveStatus[Drive].RootNode) Then
  Begin
    System.Delete(Path, 1, 3);
    IF Length(Path) > 0 Then
      Result := SearchSubDirs(DriveStatus[Drive].RootNode, UpperCase(Path))
    Else
      Result := DriveStatus[Drive].RootNode;
  End;
End; {FindNodetoPath}


Function TDriveView.IterateSubTree(Var StartNode : TTreeNode;
                                   CallBackFunc  : TCallBackFunc;
                                   Recurse       : TRecursiveScan;
                                   ScanStartNode : TScanStartNode;
                                   Data          : Pointer) : Boolean;

(* Scans StartNode and level-1 Subdirectories plus open subdirectories*)

Function ScanSubDirs(Var StartNode : TTreeNode) : Boolean;
(* Scans all subdirectories of Startnode *)
Var Node          : TTreeNode;
    NextNode      : TTreeNode;
    NodeHasChilds : Boolean;

Begin
  Result := False;
  IF Not Assigned(StartNode) Then Exit;
  Node := StartNode.GetFirstChild;

  While Assigned(Node) And FContinue Do
  Begin
    NextNode := StartNode.GetNextChild(Node);
    NodeHasChilds := Node.HasChildren;

    IF Not FContinue Or Not CallBackFunc(Node, Data) Then
    Exit;

    IF Assigned(Node) And
      ((Recurse = rsRecursive) Or
        ((Recurse = rsRecursiveExpanded) And Node.Expanded) Or
        ((Recurse = rsRecursiveExisting) And NodeHasChilds)) Then
    IF Not ScanSubDirs(Node) Or Not FContinue Then
    Exit;

    Node := NextNode;
  End;
  Result := True;
End; {ScanSubDirs}


Begin {IterateSubTree}
  Result := False;
  FContinue := True;
  IF Not Assigned(CallBackFunc) Then
  Exit;

  IF ScanStartNode = coScanStartNode Then
  CallBackFunc(StartNode, Data);

  IF Assigned(StartNode) Then
  IF Not FContinue Or Not ScanSubDirs(StartNode) Then
  Exit;

  Result := True;
End; {IterateSubTree}



Function TDriveView.CallBackValidateDir(Var Node : TTreeNode; Data: Pointer) : Boolean;

Type PSearchRec = ^TSearchRec;

Const AttrMask = faDirectory or faSysFile or faHidden;

Var WorkNode : TTreeNode;
    DelNode  : TTreeNode;
    NewNode  : TTreeNode;
    DirList  : TStringList;
    SRec     : TSearchRec;
    DosError : Integer;
    UpperDir : String;
    Index    : Integer;
    DirFound : Boolean;


Begin {CallBackValidateDir}
  Result := True;
  IF Not Assigned(Node) Or Not Assigned(Node.Data) Then
  Exit;

  {Check, if directory still exists: (but not with root directory) }
  IF Assigned(Node.Parent) And (PScanDirInfo(Data)^.StartNode = Node) Then
  IF Not DirExists(GetDirPathName(Node)) Then
  Begin
    WorkNode := Node.Parent;
    IF Selected = Node Then
    Selected := WorkNode;
    IF DropTarget = Node Then
    DropTarget := NIL;
    Node.Delete;
    Node := NIL;
    Exit;
  End;

  UpperDir := AddSlash(GetDirPath(Node));
  WorkNode := Node.GetFirstChild;

  IF Node.HasChildren And Assigned(WorkNode) Then
  {if subnodes allready exists: validate existing subnodes
   and add all new subdirectories as subnodes:}
  Begin
    DirList := TStringList.Create;
    DosError := FindFirst(UpperDir + '*.*', AttrMask, SRec);
    While DosError = 0 Do
    Begin
      IF (Srec.Name <> '.' ) And
         (Srec.Name <> '..') And
         (Srec.Attr And faDirectory <> 0) Then
      Begin
        DirList.Add(Srec.Name);

        {Does current directory exists as subnode?}
        WorkNode := Node.GetFirstChild;
        DirFound := False;
        While Assigned(WorkNode) And Not DirFound Do
        Begin
           DirFound := TNodeData(WorkNode.Data).DirName = Srec.Name;
           IF DirFound Then
           Break;
           WorkNode := Node.GetNextChild(WorkNode);
        End;

        {Subnode does not exists: add it:}
        IF Not DirFound Then
        Begin
          NewNode := AddChildNode(Node, SRec);
          TNodeData(NewNode.Data).NewDir   := True;
          PScanDirInfo(Data)^.NewDirFound  := True;
        End;
      End;
      DosError := FindNext(Srec);
    End;
    FindClose(Srec);
    Dirlist.Sort;

    {Remove not existing subnodes:}
    WorkNode := Node.GetFirstChild;
    While Assigned(WorkNode) Do
    Begin
      IF Not Assigned(WorkNode.Data) Or
         NOT DirList.Find(TNodeData(WorkNode.Data).DirName, Index) Then
      Begin
        DelNode := WorkNode;
        {NewNode := WorkNode.Parent;
        IF Selected = DelNode Then
        Selected := NewNode;}
        WorkNode := Node.GetNextChild(WorkNode);
        DelNode.Delete;
      End
      Else
      Begin
        IF (DirList[Index] <> TNodeData(WorkNode.Data).DirName) Then
        Begin
          {Case of directory letters has changed:}
          TNodeData(WorkNode.Data).DirName := DirList[Index];
          TNodeData(WorkNode.Data).ShortName := ExtractShortPathName(GetDirPathName(WorkNode));
          WorkNode.Text := DirList[Index];
        End;
        DirList.Delete(Index);
        WorkNode := Node.GetNextChild(WorkNode);
      End;
    End;
    DirList.Free;
  End
  Else
  {No subnodes allready exists: just add all found subdirectories as subnodes
   for performance reasons:}
  IF DoScanDir(Node) Then
  Begin
    DosError := FindFirst(UpperDir + '*.*', AttrMask, SRec);
    While DosError = 0 Do
    Begin
      IF (Srec.Name <> '.' ) And
         (Srec.Name <> '..') And
         (Srec.Attr And faDirectory <> 0) Then
      Begin
        NewNode := AddChildNode(Node, SRec);
        TNodeData(NewNode.Data).NewDir   := True;
        PScanDirInfo(Data)^.NewDirFound  := True;
      End;
      DosError := FindNext(Srec);

    End;
    FindClose(Srec);
  End;

  IF Node = PScanDirInfo(Data)^.StartNode Then
  IF Assigned(Node) And Assigned(Node.Data) Then
  Begin
    TNodeData(Node.Data).Verified := True;
    TNodeData(Node.Data).NewDir   := False;
  End;
  {Application.ProcessMessages;}
End; {CallBackValidateDir}



Procedure TDriveView.ValidateDirectoryEx(Node : TTreeNode; Recurse : TRecursiveScan);
Var Info     : PScanDirInfo;
    SelDir   : String;
    SaveCursor : TCursor;
    RestartWatchThread : Boolean;
    SaveCanChange      : Boolean;

Begin
  IF Not Assigned(Node)      Or
     Not Assigned(Node.Data) Or
     fValidateFlag           Or
     Not DoScanDir(Node)     Then
  Exit;

  SelDir        := Directory;
  SaveCursor    := Screen.Cursor;
  IF Self.Focused Then
  Screen.Cursor := crHourGlass;

  IF Node.Level = 0 Then
  DriveStatus[GetDrive(Node)].ChangeTimer.Enabled := False;

  RestartWatchThread := WatchThreadActive;
  Try
    IF WatchThreadActive Then
    StopWatchThread;

    fValidateFlag := True;
    New(Info);
    Info^.StartNode := Node;
    Info^.NewDirFound := False;
    SaveCanChange := FCanChange;
    FCanChange := True;
    FChangeFlag := False;
    IterateSubTree(Node, CallBackValidateDir, Recurse, coScanStartNode, Info);
    fValidateFlag := False;
    IF Not Assigned(Selected) And (Length(SelDir) > 0) Then
    Directory := Copy(SelDir, 1, 3);
    IF (SelDir <> Directory) And Not FChangeFlag Then
    Change(Selected); 
    FCanChange := SaveCanChange;

    IF Info^.NewDirFound And Assigned(Node) AND Node.HasChildren Then
    SortChildren(Node, True);

    Dispose(Info);
  Finally
    IF RestartWatchThread And fWatchDirectory And Not WatchThreadActive Then
    StartWatchThread;

    Screen.Cursor := SaveCursor;
  End;
End; {ValidateDirectoryEx}


Procedure TDriveView.ValidateDirectoryEasy(Node : TTreeNode);
Begin
  IF Not Assigned(Node) Then
  Exit;
  IF Not Assigned(Node.Data) or Not TNodeData(Node.Data).Verified Then
  ValidateDirectoryEx(Node, rsRecursiveExpanded);
End; {ValidateDirectoryEasy}


Procedure TDriveView.ValidateDirectory(Node : TTreeNode);
Begin
  ValidateDirectoryEx(Node, rsRecursiveExpanded);
End;  {ValidateDirectory}


Procedure TDriveView.ValidateExistingDirectories(Node : TTreeNode);
Begin
  ValidateDirectoryEx(Node, rsRecursiveExisting);
End; {ValidateVisibleDirectories}


Procedure TDriveView.ValidateAllDirectories(Node : TTreeNode);
Begin
  ValidateDirectoryEx(Node, rsRecursive);
End; {ValidateAllDirectories}


Function  TDriveView.GetSubTreeSize(Node : TTreeNode) : Integer;
Var PSubSize : PInt;
    SaveCursor : TCursor;
Begin
  IF Not Assigned(Node) Then
  Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['GetSubTreeSize']));
  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  ValidateAllDirectories(Node);
  RefreshDirSize(Node);
  New(PSubSize);
  PSubSize^ := 0;
  IterateSubTree(Node, CallBackSetDirSize, rsRecursive, coScanStartNode, PSubSize);
  Result := PSubSize^;
  Dispose(PSubSize);
  Screen.Cursor := SaveCursor;
End; {GetSubTreeSize}


Function TDriveView.GetDriveTypeToNode(Node : TTreeNode) : Integer;
Begin
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['GetDriveTypeToNode']));
  Result := DriveInfo[GetDirPath(Node)[1]].DriveType
End; {GetDriveTypeToNode}


Function TDriveView.GetDriveType(Drive : TDrive) : Integer;           {Returns DRIVE_CDROM etc..}
Begin
  Result := DriveInfo[UpCase(Drive)].DriveType;
End; {GetDriveType}


Function TDriveView.NodeUpdateAble(Node : TTreeNode) : Boolean;
Begin
  Result := Assigned(Node)        And
            Assigned(Node.Data)   And
            (Node.Level > 0)      And
            (GetDriveTypetoNode(Node) <> DRIVE_CDROM);
End; {NodeUpdateAble}


Function TDriveView.CallBackSaveNode(Var Node : TTreeNode; Data: Pointer) : Boolean;
Var S13 : String[13];
Begin
  Result := True;
  LogFileNode.Level     := Node.Level;
  LogFileNode.Attrs     := TNodeData(Node.Data).Attr;

  S13                   := TNodeData(Node.Data).ShortName + #0;
  Move(S13[1], LogFileNode.ShortName[0], SizeOf(LogFileNode.ShortName));

  LogFileNode.NameLen   := Length(GetDirName(Node));
  BlockWrite(TempFile, LogFileNode, SizeOf(LogFileNode));
  BlockWrite(TempFile, Node.Text[1], LogFileNode.NameLen);
End;


Function TDriveView.SaveSubNodesToFile(ParentNode : TTreeNode; FileName : String) : Boolean;
Var LogFileHeader   : TLogFileHeader;

Begin
  Result := False;
  IF Not Assigned(ParentNode) Then Exit;
  AssignFile(TempFile, FileName);
  Rewrite(TempFile, 1);

  LogFileHeader.ID := coLogFileID;
  LogFileHeader.Version := coLogFileVersion;
  BlockWrite(TempFile, LogFileHeader, SizeOf(LogFileHeader));

  IterateSubTree(ParentNode, CallbackSaveNode, rsRecursive, coNoScanStartNode, NIL);

  CloseFile(TempFile);
  Result := True;
End; {SaveSubNodesToFile}



Function TDriveView.LoadSubNodesFromFile(ParentNode : TTreeNode; FileName : String) : Boolean;
Var Name            : ShortString;
    Node            : TTreeNode;
    WorkNode        : TTreeNode;
    LogFileHeader   : TLogFileHeader;
    SRec            : TSearchRec;
    SaveCursor      : TCursor;

Begin
  Result := False;
  IF Not Assigned(ParentNode) OR Not FileExists(FileName) Then Exit;
  Node := ParentNode;
  {Items.BeginUpdate;}
  SaveCursor    := Screen.Cursor;      
  Screen.Cursor := crHourGlass;
  Try
    Node.DeleteChildren;
    AssignFile(TempFile, FileName);
    Reset(TempFile, 1);
    BlockRead(TempFile, LogFileHeader, SizeOf(LogFileHeader));
    IF (LogFileHeader.ID <> coLogFileID) Or (LogFileHeader.Version <> coLogFileVersion) Then
    Begin
      Close(TempFile);
      Exit;
    End;

    While Not EOF(TempFile) Do
    Begin
      BlockRead(TempFile, LogFileNode, SizeOf(LogFileNode));
      BlockRead(TempFile, Name[1], LogFileNode.NameLen);
      SetLength(Name, LogFileNode.NameLen);
      While Assigned(Node.Parent) And (Succ(Node.Level) > LogFileNode.Level) Do
      Begin
       {ShowMessage(Format('%u: ' + Name,[Level]));}
       Node := Node.Parent;
      End;
      Srec.Name := Name;
      Srec.Attr := LogFileNode.Attrs;
      Move(LogFileNode.ShortName, Srec.FindData.cAlternateFileName, SizeOf(LogFileNode.ShortName));
      WorkNode := AddChildNode(Node, SRec);
      Node := WorkNode;
    End;

    CloseFile(TempFile);
  Finally
    {Items.EndUpdate;}
    Screen.Cursor := SaveCursor;
  End;
  DriveStatus[GetDrive(ParentNode)].Scanned  := True;
  DriveStatus[GetDrive(ParentNode)].Verified := False;
  Result := True;
End; {LoadSubNodesFromFile}



Function  TDriveView.CallBackSaveNodeState(Var Node : TTreeNode; Data: Pointer) : Boolean;
Begin
  Result := True;
  TNodeData(Node.Data).Expanded := Node.Expanded;
End; {CallBackSaveNodeState}


Function  TDriveView.CallBackRestoreNodeState(Var Node : TTreeNode; Data: Pointer) : Boolean;
Begin
  Result := True;
  Node.Expanded := TNodeData(Node.Data).Expanded;
End; {CallBackRestoreNodeState}


Procedure TDriveView.SaveNodesState(Node : TTreeNode);
Begin
  IterateSubTree(Node, CallbackSaveNodeState, rsRecursive, coScanStartNode, NIL);
End; {SaveNodesState}


Procedure TDriveView.RestoreNodesState(Node : TTreeNode);
Begin
  Items.BeginUpdate;
  IterateSubTree(Node, CallbackRestoreNodeState, rsRecursive, coScanStartNode, NIL);
  Items.EndUpdate;
End; {RestoreNodesState}



Function TDriveView.CreateDirectory(ParentNode : TTreeNode; NewName : String) : TTreeNode;
Var IORes   : Boolean;
    Srec    : TSearchRec;

Begin
    IF Not Assigned(ParentNode) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['CreateDirectory']));

    Result := NIL;
    StopWatchThread;
    {create phyical directory:}
    MkDir(GetDirPath(ParentNode) + '\' + NewName);
    IORes := IOResult <> 0;
    StartWatchThread;
    IF Not IORes Then
    Begin
      {Create treenode:}
      FindFirst(GetDirPath(ParentNode) + '\' + NewName, faAnyFile, SRec);
      Result := AddChildNode(ParentNode, Srec);
      FindClose(Srec);
      SortChildren(ParentNode, False);
      ParentNode.Expand(False);
    End;
End; {CreateDirectory}


Function TDriveView.DeleteDirectory(Node: TTreeNode; AllowUndo : Boolean) :Boolean;
Var DelDir         : String;
    OperatorResult : Boolean;
    PrevNode       : TTreeNode;
    FileOperator   : TFileOperator;
    SaveCursor     : TCursor;

Begin
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['DeleteDirectory']));

  Result := False;
  IF Assigned(Node) And (Node.Level > 0) Then
  Begin
    SaveCursor    := Screen.Cursor;
    Screen.Cursor := crHourGlass;
    FileOperator  := TFileOperator.Create(Self);
    DelDir := GetDirPathName(Node);
    FileOperator.OperandFrom.Add(DelDir);
    FileOperator.Operation := foDelete;
    IF AllowUndo Then
    FileOperator.Flags := FileOperator.Flags + [foAllowUndo]
    Else
    FileOperator.Flags := FileOperator.Flags - [foAllowUndo];

    Try
      IF DirExists(DelDir) Then
      Begin
        StopWatchThread;
        OperatorResult := FileOperator.Execute;

        IF OperatorResult And Not FileOperator.OperationAborted And Not DirExists(DelDir) Then
        Begin
          PrevNode := Node.GetPrevVisible;
          IF Assigned(PrevNode) Then
          Directory := GetDirPathName(PrevNode);
          Node.Delete;
        End
        Else
        Begin
          Result := False;
          IF (Win32PlatForm = VER_PLATFORM_WIN32_NT) And Not AllowUndo Then
          Begin
            {WinNT4-Bug: FindFirst still returns the directories search record, even if the
             directory was deleted:}
            ChDir(DelDir);
            IF IOResult <> 0 Then
            Begin
              PrevNode := Node.GetPrevVisible;
              IF Assigned(PrevNode) Then
              Directory := GetDirPathName(PrevNode);
              Node.Delete;
            End;
          End;
        End;
      End
      Else
      Begin
        Node.Delete;
        Result := True;
      End;
    Finally
      StartWatchThread;
      IF Assigned(DirView) And Assigned(Selected) Then
      DirView.Path := GetDirPathName(Selected);
      FileOperator.Free;
      Screen.Cursor := SaveCursor;
    End;
  End;
End; {DeleteDirectory}


Procedure TDriveView.CreateWatchThread(Drive : TDrive);
Begin
  IF (csDesigning in ComponentState) Then
  Exit;
  IF Not Assigned(DriveStatus[Drive].DiscMonitor) Then
  With DriveStatus[Drive] Do
  Begin
    DiscMonitor := TDiscMonitor.Create(Self);
    DiscMonitor.ChangeDelay := msThreadChangeDelay;
    DiscMonitor.SubTree     := True;
    DiscMonitor.Filters     := [moDirName];
    DiscMonitor.OnChange    := ChangeDetected;
    DiscMonitor.OnInvalid   := ChangeInvalid;
    DiscMonitor.Directory   := Drive + ':\';
    DiscMonitor.Open;
  End;
End; {CreateWatchThread}


Procedure TDriveView.SetWatchDirectory(Watch : Boolean);
Begin
  IF fWatchDirectory <> Watch Then
  Begin
    fWatchDirectory := Watch;
    IF Not (csDesigning in ComponentState) And Watch Then
      StartAllWatchThreads
      Else
      StopAllWatchThreads;
  End;
End; {SetAutoScan}


Procedure TDriveView.SetDirView(DV : TDirView);
Begin
  IF Assigned(fDirView) Then
  fDirView.DriveView := NIL;

  fDirView := DV;

  IF Assigned(fDirView) Then
  fDirView.DriveView := Self;
End; {SetDirView}


Procedure TDriveView.SetChangeInterval(Interval : Cardinal);
Var Drive : TDrive;
Begin
  IF Interval > 0 Then
  Begin
    fChangeInterval := Interval;
    For Drive := FirstDrive To LastDrive Do
    With DriveStatus[Drive] Do
    IF Assigned(ChangeTimer) Then
    ChangeTimer.Interval := Interval;
  End;
End; {SetChangeInterval}


Procedure TDriveView.SetDimmHiddenDirs(DimmIt : Boolean);
Begin
  IF DimmIt <> fDimmHiddenDirs Then
  Begin
    fDimmHiddenDirs := DimmIt;
    Self.Invalidate;
  End;
End; {SetDimmHiddenDirs}


Procedure TDriveView.DeleteSubNodes(Node : TTreeNode);
Begin
  IF Assigned(Node) Then
  Begin
    Node.DeleteChildren;
    IF Node.Level = 0 Then
      DriveStatus[GetDrive(Node)].Scanned := False;
    Node.HasChildren := False;
  End;
End; {DeleteSubNodes}


Function TDriveView.NodeWatched(Node : TTreeNode) : Boolean;
Var Drive : TDrive;
Begin
  Drive := GetDrive(Node);
  Result := Assigned(DriveStatus[Drive].DiscMonitor) And
            DriveStatus[Drive].DiscMonitor.Active;
End; {NodeWatched}


procedure TDriveView.ChangeInvalid(Sender: TObject);
Var Dir     : String;

Begin
  Dir := (Sender as TDiscMonitor).Directory;
  With DriveStatus[Dir[1]] Do
  Begin
    DiscMonitor.Close;
    IF Assigned(fOnChangeInvalid) Then
    fOnChangeInvalid(Self, Dir[1]);
  End;
End; {DirWatchChangeInvalid}


procedure TDriveView.ChangeDetected(Sender: TObject);
Var DirChanged : String;
Begin
  IF (Sender is TDiscMonitor) Then
  Begin
    DirChanged := (Sender as TDiscMonitor).Directory;
    IF Length(DirChanged) > 0 Then
    With DriveStatus[DirChanged[1]] Do
    Begin
      ChangeTimer.Interval := 0;
      ChangeTimer.Interval := fChangeInterval;
      ChangeTimer.Enabled  := True;
    End;
  End;
End; {DirWatchChangeDetected}


Procedure TDriveView.ChangeTimerOnTimer(Sender : TObject);
Var Node  : TTreeNode;
    Drive : TDrive;
Begin
  IF Sender is TTimer Then
  With TTimer(Sender) Do
  Begin
    Drive := Chr(Tag);
    Node := FindNodeToPath(Drive + ':\');
    Interval := 0;
    Enabled  := False;

    IF Assigned(Node) Then
    Begin
{$IFDEF IDOSWIN32}
      ValidateDirectory(Node); 
{$ELSE}
      {Check also collapsed (invisible) subdirectories:}
      ValidateExistingDirectories(Node);
{$ENDIF}
      IF Assigned(fOnChangeDetected) Then
        fOnChangeDetected(Self, Drive);
    End;
  End;
End; {ChangeTimerOnTimer}


Procedure TDriveView.StartWatchThread;
Var NewWatchedDir  : String;
    Drive          : TDrive;

Begin
    IF (csDesigning in ComponentState) Or
       Not Assigned(Selected) Or
       Not fWatchDirectory Then
       Exit;

    NewWatchedDir  := GetDirPathName(RootNode(Selected));
    Drive          := Upcase(NewWatchedDir[1]);

    With DriveStatus[Drive] Do
    Begin
      IF Not Assigned(DiscMonitor) Then
      CreateWatchThread(Drive);

      IF Not DiscMonitor.Active Then
        DiscMonitor.Open;
    End;
End; {StartWatchThread}


Procedure TDriveView.StopWatchThread;
Begin
  IF Assigned(Selected) Then
  With DriveStatus[GetDrive(Selected)] Do
  IF Assigned(DiscMonitor) Then
  DiscMonitor.Close;
End; {StopWatchThread}


Procedure TDriveView.TerminateWatchThread(Drive : TDrive);
Begin
  With DriveStatus[Drive] Do
  IF Assigned(DiscMonitor) Then
  Begin
    DiscMonitor.Free;
    DiscMonitor := NIL;
  End;
End; {StopWatchThread}


Procedure TDriveView.StartAllWatchThreads;
Var Drive : TDrive;
Begin
    IF (csDesigning in ComponentState) Or
       Not FWatchDirectory Then
       Exit;

    For Drive := FirstDrive To LastDrive Do
    With DriveStatus[Drive] Do
    IF Scanned Then
    Begin
      IF Not Assigned(DiscMonitor) Then
      CreateWatchThread(Drive);
      IF Not DiscMonitor.Active Then
        DiscMonitor.Open;
    End;
End; {StartAllWatchThreads}


Procedure TDriveView.StopAllWatchThreads;
Var Drive : TDrive;
Begin
    For Drive := FirstDrive To LastDrive Do
    With DriveStatus[Drive] Do
    Begin
      IF Assigned(DiscMonitor) Then
        DiscMonitor.Close;
    End;
End; {StopAllWatchThreads}


Function TDriveView.WatchThreadActive : Boolean;
Var Drive : TDrive;
Begin
  IF Not Assigned(Selected) Then
  Begin
    Result := False;
    Exit;
  End;

  Drive := GetDrive(Selected);

  Result := FWatchDirectory        And
            Assigned(DriveStatus[Drive].DiscMonitor) And
            DriveStatus[Drive].DiscMonitor.Active;
End; {WatchThreadActive}


Procedure TDriveView.SetFullDriveScan(DoFullDriveScan : Boolean);
Begin
  IF fFullDriveScan <> DoFullDriveScan Then
  Begin
    fFullDriveScan := DoFullDriveScan;
    {IF FullDriveScan And Assigned(Selected) And Not (csDesigning in ComponentState) Then
    ValidateAllDirectories(RootNode(Selected));}
  End;
End; {SetAutoScan}


Function TDriveView.GetDirectory : String;
Begin
  IF Assigned(Selected) Then
  Result := GetDirPathName(Selected)
  Else
  Result := '';
End; {GetDirectory}


Procedure TDriveView.SetDirectory(Path : String);
Var NewSel : TTreeNode;
Begin
  FDirectory := Path;
  {Find existing path or parent path of not existing path:}
  Repeat
    NewSel := FindNodeToPath(Path);
    IF Not Assigned(NewSel) Then
    Path := ExtractFilePath(RemoveSlash(Path));
  Until Assigned(NewSel) Or (Length(Path) < 3);

  IF Assigned(NewSel) Then
  Begin
    FCanChange  := True;
    Selected := NewSel;

    IF (Selected.Level = 0) Then
    Begin
      IF Not DriveStatus[GetDrive(Selected)].Scanned Then
      ScanDrive(GetDrive(Selected));
    End;
  End
  Else
  IF csDesigning in ComponentState Then
    Selected := NIL;
End; {SetDirectory}



{Centers the Node vertically in the treeview window:}
Procedure TDriveView.CenterNode(Node : TTreeNode);
Var NextNode : TTreeNode;
    NodePos  : TRect;
    Scrolled : Boolean;

Begin
  IF Not Assigned(Node) Then
  Exit;
  Node.MakeVisible;

  NextNode := TopItem;
  NodePos  := Node.DisplayRect(False);
  Scrolled := False;
  IF (NodePos.Top < Height Div 3) Or (NodePos.Top > ((Height * 2) Div 3)) Then
  Begin
   While Assigned(NextNode) And (NodePos.Top < (Height Div 3)) Do
   Begin
     NextNode := NextNode.GetPrevVisible;
     Scrolled := True;
     IF Assigned(NextNode) Then
     Begin
       TopItem := NextNode;
       NodePos := Node.DisplayRect(False);
     End;
   End;

   IF Not Scrolled Then
   While Assigned(NextNode) And (NodePos.Top > ((Height * 2) Div 3)) Do
   Begin
     NextNode := NextNode.GetNextVisible;
     IF Assigned(NextNode) Then
     Begin
       TopItem := NextNode;
       NodePos := Node.DisplayRect(False);
     End;
   End;
  End;
End; {CenterNode}


Function TDriveView.GetDirName(Node : TTreeNode) : String;
Begin
    IF Assigned(Node) And Assigned(Node.Data) Then
      Result := TNodeData(Node.Data).Dirname
    Else
      Result := '';
End; {GetDirName}


{GetDirPath: Allways returns the complete path to Node without the trailing backslash:
 C:, C:\WINDOWS, C:\WINDOWS\SYSTEM  }
Function TDriveView.GetDirPath (Node : TTreeNode) : String;
Var T    : TTreeNode;
    PStr : String;

Begin
  IF Not Assigned(Node) Then
  Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['GetDirPath']));

  PStr := GetDirName(Node);
  T    := Node.Parent;

  While (T <> NIL) AND (T.Level >= 0) Do
  Begin
    IF T.Level > 0 Then
    PStr := GetDirName(T) + '\' + PStr
    Else
    PStr := GetDirName(T) + PStr;
    T    := T.Parent;
  End;
  IF Length(PStr) = 3 Then
    Result := Copy(PStr,1, 2)
  Else
    Result := PStr;
End; {GetDirPath}



{GetDirPathName: Returns the complete path to Node with trailing backslash on rootnodes:
 C:\ ,C:\WINDOWS, C:\WINDOWS\SYSTEM  }
Function TDriveView.GetDirPathName(Node: TTreeNode) : String;
Begin
  Result := GetDirPath(Node);
  IF Length(Result) = 2 Then
  Result := Result + '\';
End; {GetDirPathName}


{GetDrive: returns the driveletter of the Node.}
Function TDriveView.GetDrive(Node : TTreeNode) : Char;
Var Path : String;
Begin
  IF Not Assigned (Node) Or Not Assigned(Node.Data) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['GetDrive']));
  Path := GetDirPath(Node);
  IF Length(Path) > 0 Then
  Result := Upcase(Path[1])
  Else
  Result := #0;
End; {GetDrive}


{RootNode: returns the rootnode to the Node:}
Function TDriveView.RootNode(Node : TTreeNode) : TTreeNode;
Begin
  Result := Node;
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['RootNode']));
  While Assigned(Result.Parent) Do
  Result := Result.Parent;
End; {RootNode}


{NodeAttr: Returns the directory attributes to the node:}
Function TDriveView.NodeAttr(Node : TTreeNode) : Integer;
Begin
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['NodeAttr']));
  Result := TNodeData(Node.Data).Attr;
End; {NodeAttr}


Function TDriveView.NodeVerified(Node : TTreeNode) : Boolean;
Begin
  IF Not Assigned(Node) Or Not Assigned(Node.Data) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['NodeVerified']));
  Result := TNodeData(Node.Data).Verified;
End; {NodeVerified}



Procedure TDriveView.SetBoldDraw(Node : TTreeNode; BoldDraw : Boolean);
Begin
  IF Not Assigned(Node) Or Not Assigned(Node.Data) Then
     Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['SetBoldDraw']));

  IF TNodeData(Node.Data).DrawBold <> BoldDraw Then
  Begin
    TNodeData(Node.Data).DrawBold := BoldDraw;
    Node.Text := Node.Text; {Force redraw}
  End;
End; {SetBoldDraw}


Function  TDriveView.CallBackExpandLevel(Var Node : TTreeNode; Data: Pointer) : Boolean;
Begin
  Result := True;
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['CallBackExpandLevel']));
  IF (Node.Level <= Integer(Data)) And Not Node.Expanded Then
  Node.Expand(False)
  Else IF (Node.Level > Integer(Data)) And Node.Expanded Then
  Node.Collapse(True);
End; {CallBackExpandLevel}


Procedure TDriveView.ExpandLevel(Node : TTreeNode; Level : Integer);
{Purpose: Expands all subnodes of node up to the given level}
Begin
  IF Not Assigned(Node) Or Not Assigned(Node.Data) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['ExpandLevel']));
  Items.BeginUpdate;
  IterateSubTree(Node, CallBackExpandLevel, rsRecursive, coScanStartNode, Pointer(Level));
  Items.EndUpdate;
End; {ExpandLevel}


Function  TDriveView.CallBackDisplayName(Var Node : TTreeNode; Data: Pointer) : Boolean;
Begin
  Result := True;
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['CallBackDisplayName']));
  Node.Text := GetDisplayName(Node);
End; {CallBackDisplayName}


Function  TDriveView.CallBackSetDirSize(Var Node : TTreeNode; Data: Pointer) : Boolean;
Begin
  Result := True;
  IF Assigned(Node) Then
  Begin
    SetDirSize(Node);
    IF fShowDirSize Then
    Node.Text := GetDisplayName(Node);
    IF Assigned(Data) Then
    INC(PInt(Data)^, TNodeData(Node.Data).DirSize);
  End;
  Application.ProcessMessages;
  IF Not FContinue Then
  Exit;
End; {CallBackSetDirSize}




Function TDriveView.FormatDirSize(Size : Cardinal) : String;
Var FSize : Cardinal;
Begin
  FSize := Size;
  IF (Size > 0) And (Size < 1024) Then
  fSize := 1
  Else
  fSize := fSize DIV 1024;
  IF fSize <= 99999 Then
  Result := FormatSize(FSize) + 'K'
  Else
  Result := FormatSize(FSize DIV 1024) + 'M';
End; {FormatDirSize}


Procedure TDriveView.SetShowDirSize(ShowIt : Boolean);
Var Drive    : Char;
    RootNode : TTreeNode;
    SaveCursor: TCursor;

Begin
  IF ShowIt = fShowDirSize Then
  Exit;
  fShowDirSize := ShowIt;
  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourglass;
  Items.BeginUpdate;
  For Drive := FirstFixedDrive To LastDrive Do
  Begin
    IF DriveInfo[Drive].Valid Then
    Begin
      RootNode := DriveStatus[Drive].RootNode;
      IF Assigned(RootNode) Then
        IterateSubTree(RootNode, CallBackDisplayName, rsRecursive, coScanStartNode, NIL);
    End;
  End;
  Items.EndUpdate;
  Screen.Cursor := SaveCursor;
End; {SetShowDirSize}


Procedure TDriveView.RefreshDirSize(Node : TTreeNode);
Begin
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['RefreshDirSize']));
  CallBackSetDirSize(Node, NIL);
End; {RefreshDirSize}


Procedure TDriveView.RefreshDriveDirSize(Drive : TDrive);
Var SaveCursor : TCursor;
Begin
  SaveCursor    := Screen.Cursor;
  Screen.Cursor := crHourglass;
  Items.BeginUpdate;
  With DriveStatus[Drive] Do
  Begin
    IF Assigned(RootNode) Then
    IterateSubTree(RootNode, CallBackSetDirSize, rsRecursive, coScanStartNode, NIL);
  End;
  Items.EndUpdate;
  Screen.Cursor := SaveCursor;
End; {RefreshDriveDirSize}



Function TDriveView.GetDirSize(Node : TTreeNode) : Cardinal;
Begin
   IF Not Assigned(Node) Or Not Assigned(Node.Data) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['GetDirSize']));
   IF TNodeData(Node.Data).DirSize = InvalidSize Then
   SetDirSize(Node);
   Result := TNodeData(Node.Data).DirSize;
End; {GetDirSize}


Procedure TDriveView.SetDirSize(Node : TTreeNode);
Var SRec : TSearchRec;
    Size : Cardinal;
Begin
   IF Not Assigned(Node) Or Not Assigned(Node.Data) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['SetDirSize']));
   Size := 0;
   IF FindFirst(AddSlash(GetDirPath(Node)) + '*.*', faAnyFile, SRec) = 0 Then
   Begin
     Repeat
       IF (Srec.Attr And faDirectory) = 0 Then
       INC(Size, Srec.Size);
     Until FindNext(Srec) <> 0;
     FindClose(Srec);
   End;
   TNodeData(Node.Data).DirSize := Size;
End; {SetDirSize}



Function TDriveView.GetDisplayName(Node : TTreeNode) : String;
Begin
  Result := '';
  IF Not Assigned(Node) OR Not Assigned(Node.Data) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['GetDisplayName']));

  IF Node.Level = 0 Then
    Result := GetDriveText(GetDrive(Node))
  Else
    Result := GetDirName(Node);

  IF fShowDirSize Then
  Result := Result + ' = ' + FormatDirSize(GetDirSize(Node));
End; {GetDisplayName}


Procedure TDriveView.SetShowVolLabel(ShowIt : Boolean);
Begin
  IF ShowIt = fShowVolLabel Then
  Exit;
  fShowVolLabel := ShowIt;
  RefreshRootNodes(False, dvdsFloppy);
End; {SetShowVolLabel}


Procedure TDriveView.SetVolDisplayStyle(doStyle : TVolumeDisplayStyle);
Begin
  IF doStyle <> fVolDisplayStyle Then
  Begin
    fVolDisplayStyle := doStyle;
    IF Not fCreating Then
    RefreshRootNodes(False, dvdsFloppy);
  End;
End; {SetVolDisplayStyle}


Procedure TDriveView.DisplayContextMenu(Node : TTreeNode);
Var Verb              : String;

Begin
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['DisplayContextMenu']));
  IF Node <> Selected Then
  DropTarget := Node;
  {ItemProp.DisplayContextMenu(GetDirPathName(Node), Application.Handle, Mouse.CursorPos);}

  Verb := EmptyStr;
  ShellDisplayContextMenu(FParentForm.Handle, Mouse.CursorPos, GetDirPathName(Node), CanEdit(Node), Verb);
  If Verb = 'rename' Then
    Node.EditText
  Else If Verb = shcCut Then
  Begin
    LastClipBoardOperation := cboCut;
    LastPathCut := GetDirPathName(Node);
  End
  Else If Verb = shcCopy Then
    LastClipBoardOperation := cboCopy;

  DropTarget := NIL;
End; {ContextMenu}


Procedure TDriveView.DisplayPropertiesMenu(Node : TTreeNode);
Begin
  IF Not Assigned(Node) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['DisplayPropertiesMenu']));
  ShellExecuteContextCommand(FParentForm.Handle, shcProperties, GetDirPathName(Node));
End; {ContextMenu}


Function TDriveView.SortChildren(ParentNode : TTreeNode; Recurse : Boolean) : Boolean;
Var Node : TTreeNode;
Begin
  Result := False;
  IF Not Assigned(ParentNode) Then
    Raise ENodeNotAssigned.Create(Format(ErrorNodeNA, ['SortChildren']));
  IF TreeView_SortChildren(Self.Handle, ParentNode.ItemID, 0) Then
  Begin
    Result := True;
    IF Recurse Then
    Begin
      Node := ParentNode.GetFirstChild;
      While Assigned(Node) Do
      Begin
        IF Node.HasChildren Then
        SortChildren(Node, Recurse);
        Node := ParentNode.GetNextChild(Node);
      End;
    End;
  End;
End; {SortChildren}


Procedure TDriveView.Notification(AComponent: TComponent; Operation: TOperation);
begin
  Inherited Notification(AComponent, Operation);
  If (Operation = opRemove) Then
  Begin
    IF AComponent = fDirView then
    fDirView := NIL
    Else
    IF AComponent = FDriveBox then
    FDriveBox := NIL
  End;
end; {Notification}


Procedure TDriveView.SetSelected(Node : TTreeNode);
Begin
  IF Node <> Selected Then
  Begin
    FChangeFlag := False;
    FCanChange := True;
    Inherited Selected := Node;
    IF Not fChangeFlag Then
    Change(Selected);
  End;
End; {SetSelected}


{=================================================================}
{ Drag&Drop handling:                                             }
{=================================================================}
{$IFDEF DRAGDROP}

{Called by TFileDeleteThread, when a file deletion was detected by the D&D receiving application:}
Procedure TDriveView.SignalDirDelete(Sender: TObject; Files : TStringList);
Begin
  IF Files.Count > 0 Then
    ValidateDirectory(FindNodeToPath(Files[0]));
End; {SignalDirDelete}


Procedure TDriveView.DDDragEnter(DataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: longint; var Accept:boolean);
Var KeyBoardState : TKeyBoardState;
    i             : Integer;

Begin
  IF (FDD.FileList.Count > 0) And (Length(TFDDListItem(FDD.FileList[0]^).Name) > 0) Then
  Begin
    FDragDrive := TFDDListItem(FDD.FileList[0]^).Name[1];
    FExeDrag   :=  (FDD.AvailableDropEffects And DropEffect_Link) <> 0;
    IF FExeDrag Then
    For i := 0 To FDD.FileList.Count - 1 Do
      If Not isExecutable(TFDDListItem(FDD.FileList[i]^).Name) Then
      Begin
        FExeDrag := False;
        Break;
      End;
  End
  Else
    FDragDrive := #0;

  GetSystemTimeAsFileTime(DragOverTime);
  GetSystemTimeAsFileTime(LastHScrollTime);
  GetSystemTimeAsFileTime(LastVScrollTime);
  VScrollCount := 0;

  IF (GetKeyState(VK_SPACE) <> 0) And GetKeyboardState(KeyBoardState) Then
  Begin
    KeyBoardState[VK_Space] := 0;
    SetKeyBoardState(KeyBoardState);
  End;

  IF Assigned(FOnDDDragEnter) Then
  FOnDDDragEnter(Self, DataObj, grfKeyState, Pt, dwEffect, Accept);
End; {DDDragEnter}


Procedure TDriveView.DDDragLeave;
Begin
  IF Assigned(DropTarget) Then
  Begin
    IF DragImageList.Dragging Then
    Begin
      DragImageList.HideDragImage;
      DropTarget := NIL;
      DragImageList.ShowDragImage;
    End
    Else
    DropTarget := NIL;
  End;

  IF Assigned(FOnDDDragLeave) Then
  FOnDDDragLeave(Self);
End; {DragLeave}


Procedure TDriveView.DDDragOver(grfKeyState: Longint; pt: TPoint; var dwEffect: longint);
Var Node         : TTreeNode;
    KnowTime     : FILETIME;
    TempTopItem  : TTreeNode;
    NbPixels     : Integer;
    ScrollInfo   : TScrollInfo;
    KeyBoardState: TKeyBoardState;
    Rect1        : Trect;
    UpdateImage  : Boolean;
    LastDragNode : TTreeNode;
    TargetDrive  : Char;

begin
  IF dwEffect <> DropEffect_None Then
  Begin
    PT := ScreenToClient(PT);
    Node := GetNodeAt(Pt.X, Pt.Y);

    IF Assigned(Node) Then
    Begin
      LastDragNode := DropTarget;
      UpdateImage := False;
      IF DragImageList.Dragging And (LastDragNode <> Node) Then
      Begin
        IF Assigned(LastDragNode) Then
        Begin
          Rect1 := LastDragNode.DisplayRect(True);
          IF Rect1.Right >= Pt.x - DragImageList.GetHotSpot.X Then
          Begin
            DragImageList.HideDragImage;
            UpdateImage := True;
          End
          Else
          Begin
            Rect1 := Node.DisplayRect(True);
            IF Rect1.Right >= Pt.x - DragImageList.GetHotSpot.X Then
            Begin
              DragImageList.HideDragImage;
              UpdateImage := True;
            End
          End;
        End
        Else
        {LastDragNode not assigned:}
        Begin
          DragImageList.HideDragImage;
          UpdateImage := True;
        End;
      End;
      DropTarget := Node;
      IF UpdateImage Then
        DragImageList.ShowDragImage;

      TargetDrive := GetDirPath(Node)[1];

      {Drop-operation allowed at this location?}
      IF Assigned(DragNode) And
         (dwEffect <> DropEffect_Link) And
        ((Node = DragNode) Or Node.HasAsParent(DragNode) Or (DragNode.Parent = Node)) Then
        dwEffect := DropEffect_None;

      GetSystemTimeAsFileTime(KnowTime);

      IF GetKeyState(VK_SPACE) = 0 Then
      Begin
        {Expand node after 2.5 seconds: }
        IF Not Assigned(LastDragNode) Or (LastDragNode <> Node) Then
          GetSystemTimeAsFileTime(DragOverTime)     {not previous droptarget: start timer}
        Else
        Begin
          IF ((INT64(KnowTime) - INT64(DragOverTime)) > DDExpandDelay) Then
          Begin
            TempTopItem := TopItem;
            DragImageList.HideDragImage;
            Node.Expand(False);
            TopItem := TempTopItem;
            Application.ProcessMessages;
            DragImageList.ShowDragImage;
            DragOverTime := KnowTime;
          End;
        End;
      End
      Else
      Begin
        {restart timer}
        GetSystemTimeAsFileTime(DragOverTime);
        IF GetKeyboardState(KeyBoardState) Then
        Begin
          KeyBoardState[VK_Space] := 0;
          SetKeyBoardState(KeyBoardState);
        End;

        TempTopItem := TopItem;
        DragImageList.HideDragImage;
        IF Not Node.HasChildren Then
        ValidateDirectory(Node);
        IF Node.Expanded Then
        Begin
          IF Not Selected.HasAsParent(Node) Then
          Node.Collapse(False);
        End
        Else
        Node.Expand(False);
        TopItem := TempTopItem;
        Application.ProcessMessages;
        DragImageList.ShowDragImage;
      End;

      NbPixels := Abs((Font.Height));

      {Vertical treescrolling:}
      IF ((INT64(KnowTime) - INT64(LastVScrollTime)) > DDVScrollDelay) OR
         ((VScrollCount > 5) And ((INT64(KnowTime) - INT64(LastVScrollTime)) > (DDVScrollDelay Div 2))) Then
      Begin
        {Scroll tree up, if droptarget is topitem:}
        IF Node = TopItem Then
        Begin
          DragImageList.HideDragImage;
          Perform(WM_VSCROLL, SB_LINEUP, 0);
          DragImageList.ShowDragImage;
          GetSystemTimeAsFileTime(LastVScrollTime);
          INC(VScrollCount);
        End
        Else
        {Scroll tree down, if next visible item of droptarget is not visible:}
        Begin
          IF PT.Y + 3 * nbPixels > Height Then
          Begin
            DragImageList.HideDragImage;
            Perform(WM_VSCROLL, SB_LINEDOWN, 0);
            DragImageList.ShowDragImage;
            GetSystemTimeAsFileTime(LastVScrollTime);
            INC(VScrollCount);
          End
          Else
          Begin
            VScrollCount := 0;
          End;
        End;
      End; {VScrollDelay}

      {Horizontal treescrolling:}
      {Scroll tree Left}
      IF ((INT64(KnowTime) - INT64(LastHScrollTime)) > DDHScrollDelay) Then
      Begin
        GetSystemTimeAsFileTime(LastHScrollTime);
        GetScrollInfo(Handle, SB_HORZ, ScrollInfo);
        if ScrollInfo.nMin <> ScrollInfo.nMax then
        Begin
          if (PT.X < 50 ) then
          Begin
            IF Node.DisplayRect(True).Right + 50 < Width Then
            Begin
              DragImageList.HideDragImage;
              Perform(WM_HSCROLL, SB_LINELEFT, 0);
              DragImageList.ShowDragImage;
            End;
          End
          Else
          IF (PT.X > (Width - 50)) Then
          Begin
            IF Node.DisplayRect(True).Left  > 50  Then
            Begin
              DragImageList.HideDragImage;
              Perform(WM_HSCROLL, SB_LINERIGHT, 0);
              DragImageList.ShowDragImage;
            End;
          End;
        End;
      End;

    {Set Drop effect:}
    IF (GetDriveType(TargetDrive) = DRIVE_CDROM) Or
       (TNodeData(DropTarget.Data).isRecycleBin And FDD.FileNamesAreMapped) Then
      dwEffect := DropEffect_None
    Else
    Begin
      IF TNodeData(DropTarget.Data).isRecycleBin Then
      dwEffect := DropEffect_Move
      Else
      IF (grfKeyState And (MK_CONTROL Or MK_SHIFT) = 0) Then
      Begin
        If FExeDrag And (TargetDrive >= FirstFixedDrive) Then
          dwEffect := DropEffect_Link
        Else
        IF (dwEffect = DropEffect_Copy) And
           ((DragDrive = GetDrive(DropTarget)) And
             (FDD.AvailableDropEffects and DropEffect_Move <> 0)) Then
             dwEffect := DropEffect_Move;
      End;
    End;



    End {Assigned(Node)}
    Else
    dwEffect := DropEffect_None;
  End;


  IF Assigned(FOnDDDragOver) Then
  FOnDDDragOver(Self, grfKeyState, Pt, dwEffect);
End; {DDDragOver}


Procedure TDriveView.DDDrop(DataObj: IDataObject; grfKeyState: Longint;  pt: TPoint; var dwEffect: longint);
Begin
  IF DragImageList.Dragging Then
  DragImageList.HideDragImage;
  IF dwEffect = DropEffect_None Then
  DropTarget := NIL;
  IF Assigned(FOnDDDrop) Then
  FOnDDDrop(Self, DataObj, grfKeyState, Pt, dwEffect);
End; {DDDrop}


Procedure TDriveView.DDQueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint; var Result: HResult);
Var P       : TPoint;
    ClientP : TPoint;

Begin
  IF Assigned(FOnDDQueryContinueDrag) Then
  FOnDDQueryContinueDrag(Self, fEscapePressed, grfKeyState, Result);

  IF fEscapePressed Then
  Begin
    IF DragImageList.Dragging Then
      DragImageList.HideDragImage;
    DropTarget := NIL;
    Exit;
  End;

  IF DragImageList.Dragging Then
  Begin
    GetCursorPos(P);
    {Convert screen coordinates to the parentforms coordinates:}
    ClientP := FParentForm.ScreenToClient(P);
    {Move the drag image to the new position and show it:}
    IF Not CompareMem(@ClientP, @FDragPos, SizeOf(TPoint)) Then
    Begin
      FDragPos := ClientP;
      IF PtInRect(FParentForm.BoundsRect, P) Then
      Begin
        DragImageList.DragMove(ClientP.X, ClientP.Y);
        DragImageList.ShowDragImage;
      End
      Else
        DragImageList.HideDragImage;
    End;
  End;
End; {DDQueryContinueDrag}


Procedure TDriveView.DDGiveFeedback(dwEffect: Longint; var Result: HResult);
Begin
  IF Assigned(FOnDDGiveFeedback) Then
  FOnDDGiveFeedback(Self, dwEffect, Result);
End; {DDGiveFeedback}


Procedure TDriveView.DDDragDetect(grfKeyState: Longint; DetectStart, Pt: TPoint; DragStatus:TDragDetectStatus);
Var DDResult       : TDragResult;
    DragPath       : String;
    DragParentPath : String;
    DragNodeLevel  : Integer;
    WatchThreadOK  : Boolean;
    P              : TPoint;
    Himl           : HImageList;
    NodeRect       : TRect;

Begin
  IF (DragStatus = ddsDrag) And Not Assigned(DragNode) Then
  Begin
    P := ScreenToClient(FStartPos);
    DragNode := GetNodeAt(P.X, P.Y);
  End;

  IF Assigned(FOnDDDragDetect) Then
  FOnDDDragDetect(Self, grfKeyState, DetectStart, Pt, DragStatus);

  IF (DragStatus = ddsDrag) And Assigned(DragNode) Then
  Begin
    NodeRect := DragNode.DisplayRect(True);
    Dec(NodeRect.Left, 16);

    {Check, wether the mouse cursor was within the nodes display rectangle:}
    IF (NodeRect.Left > P.X) Or (NodeRect.Right < P.X) Then
    Begin
      DragNode := NIL;
      Exit;
    End;

    FDragDrive := #0;
    {Create the dragimage:}
    IF UseDragImages Then
    Begin
      {Hide the selection mark to get a proper dragimage:}
      IF Selected = DragNode Then
      Selected := NIL;
      HIml := TreeView_CreateDragImage(Handle, DragNode.ItemID);

      {Show the selection mark if it was hidden:}
      IF Not Assigned(Selected) Then
      Selected := DragNode;

      IF Himl <> Invalid_Handle_Value Then
      Begin
         DragImageList.Handle := Himl;
         DragImageList.SetDragImage(0, P.X - NodeRect.TopLeft.X, P.Y - NodeRect.TopLeft.Y);
         P := FParentForm.ScreenToClient(Pt);
         DragImageList.BeginDrag(FParentForm.Handle, P.X, P.Y);
         DragImageList.DragLock(FParentForm.Handle, P.X, P.Y);
         ShowCursor(True);
      End;
    End;
    Dragpath := GetDirPathName(DragNode);
    IF Assigned(DragNode.Parent) Then
    DragParentPath := GetDirPathName(DragNode.Parent)
    Else
    DragParentPath := DragPath;

    DragNodeLevel := DragNode.Level;

    FDD.FileList.Clear;
    FDD.FileList.AddItem(nil,DragPath);
    IF DragNodeLevel = 0 Then
    FDD.SourceEffects := FDD.SourceEffects - [deCopy, deMove]
    Else
    FDD.SourceEffects := FDD.SourceEffects + [deCopy, deMove];

    IF GetDriveType(DragPath[1]) = DRIVE_CDROM Then
    FDD.SourceEffects := FDD.SourceEffects - [deMove];

    WatchThreadOK := WatchThreadActive;
    DropSource := Self;

    GetSystemTimeAsFileTime(DragStartTime);

    {Supress the context menu:}
    fContextMenu := False;

    {Execute the drag&drop-Operation:}
    DDResult := FDD.Execute;

    {the drag&drop operation is finished, so clean up the used drag image:}
    DragImageList.EndDrag;
    DragImageList.DragUnlock;

    Application.ProcessMessages;
    FDragDrive := #0;
    IF DDResult = drCancelled Then
      DropTarget := NIL;

    IF  (DDResult = drMove) And Not WatchThreadOK Then
    Begin
      IF (DragNodeLevel > 0) OR
        (DragParentPath <> GetDirPathName(Selected.Parent)) Then
        Begin
          DragNode := FindNodeToPath(DragPath);
          IF Assigned(DragNode) Then
          Begin
            DragFileList.Clear;
            DragFileList.Add(DragPath);
            TFileDeleteThread.Create(DragFileList, MaxWaitTimeOut, SignalDirDelete);
          End;
        End;
    End;
  DragNode := NIL;
  DropSource := NIL;
  End;
End; {(DDDragDetect}


Procedure TDriveView.DDProcessDropped(Sender: TObject; grfKeyState: Longint;  pt: TPoint; dwEffect: Longint);
Var TargetPath : String;
    KnowTime   : FILETIME;

Begin
  IF Assigned(DropTarget) Then
  Begin
    GetSystemTimeAsFileTime(KnowTime);
    TargetPath := GetDirPathName(DropTarget);
    IF DirExists(TargetPath) Then
    Begin
      IF ((INT64(KnowTime) - INT64(DragStartTime)) > DDDragStartDelay) Then
      Begin
        IF Assigned(FOnDDProcessDropped) Then
          FOnDDProcessDropped(Self, grfKeyState, pt, dwEffect);
        PerformDragDropFileOperation(TargetPath, dwEffect, TNodeData(DropTarget.Data).isRecycleBin);
        IF Assigned(FOnDDExecuted) Then
        FOnDDExecuted(Self, dwEffect);
      End;
    End
    Else
    Begin
      ValidateDirectory(DropTarget);
      DDError(DDPathNotFoundError);
    End;

    DropTarget := NIL;
    FDD.FileList.Clear;
  End;
End; {ProcessDropped}


Procedure TDriveView.PerformDragDropFileOperation(TargetPath : String; dwEffect: Integer; isRecycleBin : Boolean);
Var i                 : Integer;
    SourcePath        : String;
    SourceParentPath  : String;
    SourceFile        : String;
    SaveCursor        : TCursor;
    DoFileOperation   : Boolean;
    TargetNode        : TTreeNode;

Begin {DragDropExec}
  IF FDD.FileList.Count = 0 Then
  Exit;

  SaveCursor    := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  SourcePath    := EmptyStr;

  Try
    IF (dwEffect = DropEffect_Copy) Or
       (dwEffect = DropEffect_Move) Then
    Begin
      StopAllWatchThreads;
      If Assigned(FDirView) Then
      FDirView.StopWatchThread;

      IF Assigned(DropSource) And (DropSource is TDirView) And (DropSource <> FDirView) Then
      TDirView(DropSource).StopWatchThread;

      {Set the source directory:}
      For i := 0 to FDD.FileList.Count - 1 Do
      Begin
        FFileOperator.OperandFrom.Add(TFDDListItem(FDD.FileList[i]^).Name);
        IF FDD.FileNamesAreMapped Then
        FFileOperator.OperandTo.Add(AddSlash(TargetPath) + TFDDListItem(FDD.FileList[i]^).MappedName);
      End;

      SourcePath := TFDDListItem(FDD.FileList[0]^).Name;
      SourceParentPath := ExtractFilePath(RemoveSlash(SourcePath));

      FDD.FileList.Clear;

      FFileOperator.Flags := [foAllowUndo, foNoConfirmMkDir];

      {Set the target directory or target files:}
      IF FDD.FileNamesAreMapped And Not isRecycleBin Then
        FFileOperator.Flags := FFileOperator.Flags + [foMultiDestFiles]
      Else
      Begin
        FFileOperator.Flags := FFileOperator.Flags - [foMultiDestFiles];
        FFileOperator.OperandTo.Clear;
        FFileOperator.OperandTo.Add(TargetPath);
      End;

      IF isRecycleBin Then
        FFileOperator.Operation := foDelete
      Else
      Case dwEffect Of
        DropEffect_Copy : FFileOperator.Operation := foCopy;
        DropEffect_Move : FFileOperator.Operation := foMove;
      End;  {Case}

      IF isRecycleBin Then
      Begin
        IF Not ConfirmDelete Then
        FFileOperator.Flags := FFileOperator.Flags + [foNoConfirmation];
      End
      Else
      IF Not ConfirmOverwrite Then
      FFileOperator.Flags := FFileOperator.Flags + [foNoConfirmation];

      DoFileOperation := True;
      IF Assigned(FOnDDFileOperation) Then
      FOnDDFileOperation(Self, dwEffect, SourcePath, TargetPath, DoFileOperation);

      IF DoFileOperation And (FFileOperator.OperandFrom.Count > 0) Then
      Begin
        FFileOperator.Execute;
        IF Assigned(FOnDDFileOperationExecuted) Then
        FOnDDFileOperationExecuted(Self, dwEffect, SourcePath, TargetPath);
        IF FDD.FileNamesAreMapped Then
        FFileOperator.ClearUndo;
      End;
    End
    Else
    IF (dwEffect = DropEffect_Link) Then
    { Create Link requested: }
    Begin
      For i := 0 to FDD.FileList.Count - 1 Do
      Begin
        SourceFile := TFDDListItem(FDD.FileList[i]^).Name;
        IF Length(SourceFile) = 3 Then
        SourcePath :=  Copy(DriveInfo[SourceFile[1]].PrettyName, 4, 255) + ' (' + SourceFile[1] + ')'
        Else
        SourcePath := ExtractFileName(SourceFile);

        IF Not CreateFileShortCut(SourceFile, AddSlash(TargetPath) + SourcePath + '.lnk',
                                  ExtractFileNameOnly(SourceFile)) Then
          DDError(DDCreateShortCutError);
      End;
    End;

    IF dwEffect = DropEffect_Move Then
    Items.BeginUpdate;

    {Update source directory, if move-operation was performed:}
    IF ((dwEffect = DropEffect_Move) OR isRecycleBin) Then
      ValidateDirectory(FindNodeToPath(SourceParentPath));

    {Update subdirectories of target directory:}
    TargetNode := FindNodeToPath(TargetPath);
    IF Assigned(TargetNode) Then
    ValidateDirectory(TargetNode)
    Else
    ValidateDirectory(DriveStatus[TargetPath[1]].RootNode);

    IF dwEffect = DropEffect_Move Then
    Items.EndUpdate;

    {Update linked component TDirView:}
    IF Assigned(FDirView) And Not FDirView.WatchThreadActive Then  
      Case dwEffect of
        DropEffect_Copy,
        DropEffect_Link: If (AddSlash(TargetPath) = AddSlash(DirView.Path)) Then
                           FDirView.Reload(True);
        DropEffect_Move: If (AddSlash(TargetPath)       = AddSlash(DirView.Path)) Or
                            (AddSlash(SourceParentPath) = AddSlash(DirView.Path)) Then
                         Begin
                           IF FDirView <> DropSource Then
                           FDirView.Reload(True);
                         End;
      End; {Case}

    {Update the DropSource control, if files are moved and it is a TDirView:}
    IF (dwEffect = DropEffect_Move) And (DropSource is TDirView) Then
    TDirView(DropSource).ValidateSelectedFiles;

  Finally
    FFileOperator.OperandFrom.Clear;
    FFileOperator.OperandTo.Clear;
    StartAllWatchThreads;
    IF Assigned(FDirView) And Not FDirView.WatchThreadActive Then
    FDirView.StartWatchThread;
    IF Assigned(DropSource) And (DropSource is TDirView) And Not TDirView(DropSource).WatchThreadActive Then
    TDirView(DropSource).StartWatchThread;
    Screen.Cursor := SaveCursor;
  End;
End; {PerformDragDropFileOperation}


Procedure TDriveView.DDError(ErrorNo : TDDError);
Begin
  IF Assigned(FOnDDError) Then
  FOnDDError(Self, ErrorNo)
  Else
  Raise EDragDrop.Create('DragDrop Error: ' + IntToStr(Ord(ErrorNo)));
End; {DDError}


Function TDriveView.GetCanUndoCopyMove : Boolean;
Begin
  Result := Assigned(FFileOperator) And FFileOperator.CanUndo;
End; {CanUndoCopyMove}


Function TDriveView.UndoCopyMove : Boolean;
Var LastTarget : String;
    LastSource : String;

Begin
  Result := False;
  IF FFileOperator.CanUndo Then
  Begin
    Lasttarget := FFileOperator.LastOperandTo[0];
    LastSource := FFileOperator.LastOperandFrom[0];
    Result := FFileOperator.UndoExecute;

    IF Not WatchThreadActive Or (GetDrive(Selected) <> LastTarget[1]) Then
      ValidateDirectory(GetDriveStatus(LastTarget[1]).RootNode);
    IF LastTarget[1] <> LastSource[1] Then
      ValidateDirectory(GetDriveStatus(LastSource[1]).RootNode);

    IF Assigned(FDirView) Then
    With FDirView Do
    IF Not WatchThreadActive Then
    Begin
      IF (AddSlash(ExtractFilePath(LastTarget)) = AddSlash(Path)) Or
         (AddSlash(ExtractFilePath(LastSource)) = AddSlash(Path)) Then
        Reload(True);
    End;
  End;
End; {UndoCopyMove}


{Clipboard operations:}
Procedure TDriveView.SetLastPathCut(Path : String);
Var Node : TTreeNode;
Begin
  If FLastPathCut <> Path Then
  Begin
    Node := FindNodeToPath(FLastPathCut);
    IF Assigned(Node) Then
    Begin
      FLastPathCut := Path;
      Node.Cut     := False;
    End;
    Node := FindNodeToPath(Path);
    IF Assigned(Node) Then
    Begin
      FLastPathCut := Path;
      Node.Cut     := True;
    End;
  End;
End; {SetLastNodeCut}


Procedure TDriveView.EmptyClipboard;
Begin
  IF Windows.OpenClipBoard(0) Then
  Begin
    Windows.EmptyClipBoard;
    Windows.CloseClipBoard;
    LastPathCut := '';
    LastClipBoardOperation := cboNone;
    IF Assigned(FDirView) Then
    FDirView.EmptyClipboard;
  End;
End; {EmptyClipBoard}


Function TDriveView.CopyToClipBoard(Node : TTreeNode) : Boolean;
Begin
  Result := Assigned(Selected);
  IF Result Then
  Begin
    EmptyClipBoard;
    FDD.FileList.Clear;
    FDD.FileList.AddItem(NIL, GetDirPathName(Selected));
    Result := FDD.CopyToClipBoard;
    LastClipBoardOperation := cboCopy;
  End;
End; {CopyToClipBoard}


Function TDriveView.CutToClipBoard(Node : TTreeNode)  : Boolean;
Begin
  Result := Assigned(Node) And (Node.Level > 0) And CopyToClipBoard(Node);
  IF Result Then
  Begin
    LastPathCut := GetDirPathName(Node);
    LastClipBoardOperation := cboCut;
  End;
End; {CutToClipBoard}


Function TDriveView.CanPasteFromClipBoard : Boolean;
Begin
  Result := False;
  IF Assigned(Selected) And Windows.OpenClipboard(0) Then
  Begin
    Result := IsClipboardFormatAvailable(CF_HDROP);
    Windows.CloseClipBoard;
  End;
End; {CanPasteFromClipBoard}


Function TDriveView.PasteFromClipBoard  : Boolean;
Begin
  FDD.FileList.Clear;
  Result := False;
  IF CanPasteFromClipBoard And FDD.GetFromClipBoard Then
  Begin
    Case LastClipBoardOperation Of
      cboCopy,
      cboNone: Begin
                 PerformDragDropFileOperation(GetDirPathName(Selected), DropEffect_Copy, TNodeData(Selected.Data).isRecycleBin);
                 IF Assigned(FOnDDExecuted) Then
                 FOnDDExecuted(Self, DropEffect_Copy);
               End;
      cboCut : Begin
                 PerformDragDropFileOperation(GetDirPathName(Selected), DropEffect_Move, TNodeData(Selected.Data).isRecycleBin);
                 IF Assigned(FOnDDExecuted) Then
                 FOnDDExecuted(Self, DropEffect_Move);
                 EmptyClipBoard;
               End;
    End;
    Result := True;
  End;
End; {PasteFromClipBoard}



Procedure TDriveView.SetTargetPopUpMenu(PopMe : Boolean);
Begin
  IF PopMe <> FTargetPopUpMenu Then
  Begin
    FTargetPopUpMenu := PopMe;
    IF Assigned(FDD) Then
    FDD.TargetPopupMenu := PopMe;
  End;
End; {SetTargetPopUpMenu}

{$ENDIF} {DRAGDROP}


// ======================================================
// Initialization
// ======================================================
Initialization


  {Set language specific messagetext:}
  Case SysLocale.PriLangID Of
    LANG_GERMAN: ErrorInvalidDirName := German_ErrorInvalidDirName;
    LANG_FRENCH: ErrorInvalidDirName := French_ErrorInvalidDirName;
    Else         ErrorInvalidDirName := English_ErrorInvalidDirName;
  End; {Case}
end.
